/*
 DSPAM
 COPYRIGHT (C) 2003 NETWORK DWEEBS CORPORATION

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/stat.h>
#include <db.h>


#ifdef TIME_WITH_SYS_TIME
#   include <sys/time.h>
#   include <time.h>
#else
#   ifdef HAVE_SYS_TIME_H
#       include <sys/time.h>
#   else
#       include <time.h>
#   endif
#endif

#include "libdspam.h" 
#include "nodetree.h"
#include "dspam.h"
#include "language.h"
#include "buffer.h"
#include "storage_driver.h"
#include "error.h"
#include "config.h"
#include "util.h"
#include "decode.h"
#include "base64.h"

int main(int argc, char *argv[]) {
  char mailer_args[256];		/* Args passed to MLOCAL */
  char spam_args[256];			/* Args passed for spam */
  int mode = DSM_PROCESS;		/* Processing mode */
#ifdef TRUSTED_USER_SECURITY
  int trusted_user = 0;			/* Trusted User */
#endif

  struct nt *users;			/* Destination users */
  struct nt_node *node_nt;
  struct nt_c c_nt;

  buffer *message;			/* Message */
  char buff[1024];			/* Read Buffer */

  FILE *file;
  char filename[MAX_FILENAME_LENGTH];

#ifdef DEBUG
  char debug_args[1024];
#endif

  int user_flag = 0;
  int corpus = 0;			/* Corpus message? */
  int i;

  struct passwd *p;

  p = getpwuid(getuid());
  srand(getpid());

  spam_args[0] = 0;
  umask(006);	/* rw-rw---- */
  setbuf(stdout, NULL);

  LOGDEBUG("-- DSPAM Agent Startup --");

  users		= nt_create(NT_CHAR);
  message	= buffer_create(NULL);

  if (users == NULL || message == NULL) {
    report_error(ERROR_MEM_ALLOC);
    exit(EXIT_FAILURE);
  }

  snprintf(mailer_args, sizeof(mailer_args), "%s ", MLOCAL);

#ifdef TRUSTED_USER_SECURITY
  snprintf(filename, MAX_FILENAME_LENGTH, "%s/trusted.users", USERDIR);
  file = fopen(filename, "r");
  if (file != NULL) {
    while(fgets(buff, sizeof(buff), file)!=NULL) {
      chomp(buff);
      if (!strcmp(buff, p->pw_name)) 
        trusted_user = 1; 
    }
    fclose(file);
  } else {
    LOG(LOG_CRIT, "unable to open %s/trusted.users for reading: %s.", USERDIR, strerror(errno));
  }

  if (!trusted_user) {
    LOGDEBUG("WARNING! untrusted user.  using preconfigured parameters");
    LOGDEBUG("forcing username to %s", p->pw_name);
    nt_add(users, p->pw_name);
  }
#endif

#ifdef DEBUG
  debug_args[0] = 0;
#endif

  /* Record all pass-thru arguments passed to dspam; set modes */
  for(i=0;i<argc;i++) {

#ifdef DEBUG
    strlcat(debug_args, argv[i], sizeof(debug_args));
    strlcat(debug_args, " ", sizeof(debug_args));
#endif

#ifdef TRUSTED_USER_SECURITY
    if (!trusted_user && argv[i] != NULL &&
         (strchr(argv[i], '|') ||  strchr(argv[i], ';') ||
          strchr(argv[i], '>') || strchr(argv[i], '<') ||
          strchr(argv[i], '`'))) {
      LOG(LOG_CRIT, "insecure parameter '%s' ignored", argv[i]);
      continue;
    }
#endif

    if (!strcmp(argv[i], "--user")) {
      user_flag = 1;
      continue;
    }
    
    if (user_flag && argv[i][0] == '-') 
      user_flag = 0;
 
    if (user_flag) {
      if (
#ifdef TRUSTED_USER_SECURITY
      trusted_user &&
#endif
      argv[i]!= NULL && strlen(argv[i])<=MAX_USERNAME_LENGTH) {

#ifdef DEBUG
          LOGDEBUG("adding user: %s", argv[i]);
#endif
          nt_add(users, argv[i]);
      }
    }

    if (!strcmp(argv[i], "--version")) {
      printf("\nDSPAM Anti-Spam Suite %s (agent/library)\n\n", VERSION);
      printf("Copyright (c) 2003 Network Dweebs Corporation\n");
      printf("http://www.nuclearelephant.com/projects/dspam/\n\n");
      printf("DSPAM may be copied only under the terms of the GNU General Public License,\n");
      printf("a copy of which can be found with the DSPAM distribution kit.\n\n");
      exit(EXIT_SUCCESS);
    }

    /* --addspam tells us this has been identified as a SPAM by the user, and
         we're going to re-learn it as spam, rather than evaluate it */

    if (!strcmp(argv[i], "--addspam")) { 
      LOGDEBUG("setting mode to DSM_ADDSPAM")
      mode = DSM_ADDSPAM;
      continue;
    }

    if (!strcmp(argv[i], "--corpus")) {
      corpus = 1;  
      continue;
    }

#ifdef QLOCAL
    snprintf(spam_args, sizeof(spam_args), "%s ", QLOCAL);
    i++;
    continue;
#endif

    /* --falsepositive tells us that this has been identified as a false 
         positive by the user and we're going to re-learn it */
    if (!strcmp(argv[i], "--falsepositive")) { 
      LOGDEBUG("setting mode to DSM_FALSEPOSITIVE")
      mode = DSM_FALSEPOSITIVE;
      continue;
    }

    if (i>0 && !user_flag) {
      if (argv[i][0] == 0)
        strlcat(mailer_args, "\"\"", sizeof(mailer_args));
      else
        strlcat(mailer_args, argv[i], sizeof(mailer_args));
      strlcat(mailer_args, " ", sizeof(mailer_args));
    }
  }

#ifdef DEBUG
  LOGDEBUG("full arg list: %s", debug_args);
#endif

  if (users->items==0) {
    report_error(ERROR_USER_UNDEFINED);
    fprintf(stderr, "%s\n", SYNTAX);
    exit(EXIT_FAILURE);
  } 

#ifdef TRUSTED_USER_SECURITY
  if (!trusted_user) {
    snprintf(filename, MAX_FILENAME_LENGTH, "%s/untrusted.mailer_args", USERDIR);
    file = fopen(filename, "r");
    if (file != NULL) {
      if (fgets(buff, sizeof(buff), file)!=NULL) {
        chomp (buff);
        strlcpy(mailer_args, buff, sizeof(mailer_args));
      }
      fclose(file);
    } else {
      LOG(LOG_WARNING, "unable to locate %s/untrusted.mailer_args: %s.  allowing untrusted user to specify passthru delivery parameters", USERDIR, strerror(errno));
    }
  }
#endif

  LOGDEBUG("pass-thru arguments: %s", mailer_args);
  LOGDEBUG("identified %d destination user(s)", users->items);

  /* Read in the message */
  while((fgets(buff, sizeof(buffer), stdin))!=NULL) {
    if (buffer_cat(message, buff)) {
      LOG(LOG_CRIT, ERROR_MEM_ALLOC);
      exit(EXIT_FAILURE);
    }
  }

  if (!message->used) {
    LOG(LOG_INFO, "empty message (no data received at all)");
    exit(EXIT_SUCCESS);
  }

  /* Process message for each user */
  node_nt = c_nt_first(users, &c_nt); 
  while(node_nt != NULL) {
    struct stat s;
    char filename[MAX_FILENAME_LENGTH];
    int result;

#ifdef DEBUG
    LOGDEBUG("processing user: %s", (const char *) node_nt->ptr);
#endif

#ifdef OPT_IN
    strcpy(filename, _ds_userdir_path(node_nt->ptr, "dspam"));
#else
    strcpy(filename, _ds_userdir_path(node_nt->ptr, "nodspam"));
#endif

#ifdef OPT_IN
    if (stat(filename, &s)) {
#else
    if (!stat(filename, &s)) {
#endif
      if (!corpus)
        deliver_message(message->data, mailer_args, node_nt->ptr, mode);
    } else {
      result = process_message(message, node_nt->ptr, mode, corpus);
      if (result != DSR_ISSPAM) {
        if (result != DSR_ISINNOCENT)
          LOG(LOG_WARNING, "process_message returned error %d.  delivering message.", result);
        LOGDEBUG("delivering message");
        if (!corpus)
          deliver_message(message->data, mailer_args, node_nt->ptr,  mode);
      } else {
        LOGDEBUG("quarantining message");
        if (mode != DSM_ADDSPAM && !corpus) {
          if (spam_args[0] != 0) {
              deliver_message(message->data, spam_args, node_nt->ptr, mode);
          } else {
            if (!corpus)
#ifdef DELIVER_SPAM
              deliver_message(message->data, mailer_args, node_nt->ptr,  mode);
#else
              quarantine_message(message->data, node_nt->ptr);
#endif
          }
        }
      }
    }
    node_nt = c_nt_next(users, &c_nt);
  }

  LOGDEBUG("-- DSPAM Agent Shutdown --");
  exit(EXIT_SUCCESS);
}

/*
   process_message: calls the dspam processor for each destination user. 
                    manages signatures.
           returns: DSR_ISINNOCENT - Send message
                    DSR_ISSPAM     - Quarantine message
                    (other)        - Error
*/

int process_message(buffer *message, const char *username, int mode, int corpus) {
  DSPAM_CTX *CTX;			/* dspam context */
  struct _ds_message *components;
  struct _ds_spam_signature SIG;	/* signature object */
  char ctx_group[128];

  struct nt_node *node_nt;
  struct nt_c c_nt;

  FILE *file;
  char filename[MAX_FILENAME_LENGTH];
  char session[64];
  char *copyback;

  int have_signature = 0;
  int f_corpus = 0;
  int result;
  int i = 0;

  ctx_group[0] = 0;

#ifdef DEBUG
  {
    FILE *f;
    f = fopen("/tmp/dspam.messages", "a");
    if (f != NULL) {
      fprintf(f, "%s\n", message->data);
      fclose(f);
    }
  } 
#endif 

   snprintf(filename, sizeof(filename), "%s/group", USERDIR);
   file = fopen(filename, "r");
   if (file != NULL) {
     char *group;
     char *user;
     char buffer[1024];
     int found = 0;

     while(fgets(buffer, sizeof(buffer), file)!=NULL && found == 0) {
       chomp(buffer);
       group = strtok(buffer, ":");
       if (group != NULL) {
         user = strtok(NULL, ",");
         while(user != NULL) {
           if (!strcmp(user, username)) { 
             strlcpy(ctx_group, group, sizeof(ctx_group));
             LOGDEBUG("assigning user %s to group %s", username, group);
             found = 1;
           }
           user = strtok(NULL, ",");
         }
       }
     }
     fclose(file);
   }

   if (corpus)
     f_corpus = DSF_CORPUS;

#ifdef CHAINED_TOKENS
   CTX = dspam_init(username, ctx_group, mode, DSF_CHAINED | DSF_SIGNATURE | f_corpus);
#else
   CTX = dspam_init(username, ctx_group, mode, DSF_SIGNATURE | f_corpus);
#endif

  if (CTX == NULL) {
    LOG(LOG_WARNING, "unable to initialize dspam context");
    return EUNKNOWN;
  }

  components = _ds_actualize_message(message->data);
  if (components == NULL) {
    LOG(LOG_WARNING, "_ds_actualize_message() failed.  unable to process.");
    return EUNKNOWN;
  }

  CTX->message = components;

  /* Find, parse, and strip DSPAM signature */
  {
    char signature_key[128];
    struct nt_node *node_nt, *prev_node = NULL;
    struct nt_c c;
    struct _ds_message_block *block = NULL;
    char first_boundary[512];

#ifdef SIGNATURE_ATTACHMENTS
    int final_boundary = 0;
#else
    char *signature_begin, *signature_end, *erase_begin;
    int signature_length;
    struct nt_node *node_header;
    struct nt_c c2;
#endif

    i = 0;
    first_boundary[0] = 0;

    /* Iterate through each message component in search of a signature
       Decode components as necessary */

    node_nt = c_nt_first(CTX->message->components, &c);
    while(node_nt != NULL) {
      block = (struct _ds_message_block *) node_nt->ptr;
#ifdef VERBOSE
      LOGDEBUG("Scanning component %d for a DSPAM signature", i);
#endif

/* mode: signature-attachments */

#ifdef SIGNATURE_ATTACHMENTS
      /* record the message's main boundary so we can make sure there isn't any junk in the last component */
      if (i == 0 && block->boundary != NULL)
        strlcpy(first_boundary, block->boundary, sizeof(first_boundary));

      /* strip out any junk block that came after the end of the message */

      if (final_boundary) {
        struct nt_node *old;
        _ds_destroy_block(block);
        if (prev_node != NULL) 
          prev_node->next = node_nt->next;
        else
          components->components->first = node_nt->next;
         
        old = node_nt;
        node_nt = node_nt->next;
        free(old);
        continue;
      }

      /* if this block terminates the message, set our final boundary */
      if (block->terminating_boundary != NULL && first_boundary[0] != 0) {
        char boundary[512];
        snprintf(boundary, sizeof(boundary), "%s--", first_boundary);
        if (!strcmp(block->terminating_boundary, boundary))
          final_boundary = 1;
      }      

      /* Signature found in attachment; reassemble into _ds_spam_signature */
      if (block->media_type == MT_APPLICATION && block->media_subtype == MST_DSPAM_SIGNATURE) {
        char *token, *attachment, *m_attachment;
        long num_tokens, i = 0;
        unsigned long long value;
     
        attachment = strdup(block->body->data);
        m_attachment = attachment;
        token = strsep(&attachment, "\n");
        num_tokens = strtol(token, NULL, 0); 
        LOGDEBUG("found dspam.dat; %ld tokens to reverse", num_tokens);

        SIG.length = num_tokens * sizeof(unsigned long long);
        SIG.data = malloc(num_tokens * sizeof(unsigned long long));

        if (SIG.data == NULL) {
          LOG(LOG_CRIT, ERROR_MEM_ALLOC);
          return EUNKNOWN;
        }

        token = strsep(&attachment, "\n");
        while(token != NULL && i<num_tokens) {
          value = strtoull(token, NULL, 16);
          if (value>0) {
            memcpy(SIG.data+(i*sizeof(unsigned long long)), &value, sizeof(unsigned long long));
          }
          token = strsep(&attachment, "\n"); 
          i++;
        }
 
        CTX->signature = &SIG;
        have_signature = 2;
        free(m_attachment); 

        /* Strip out dspam data attachment */
        if (prev_node != NULL) { 
          struct _ds_message_block *prev_block = (struct _ds_message_block *) prev_node->ptr;
 
          /* use the current block's terminating boundary, since we're removing it */
          free(prev_block->terminating_boundary);
          prev_block->terminating_boundary = block->terminating_boundary;
          block->terminating_boundary = NULL;

          prev_node->next = node_nt->next;
          node_nt = prev_node;
        }
        else { 
          CTX->message->components->first = node_nt->next;
          node_nt = NULL;
        }

        _ds_destroy_block(block);

      }
#else
      if (block->media_type == MT_TEXT || block->media_type == MT_MESSAGE || i == 0) {
        char *body;

        /* Verbose output of each message component */
#ifdef VERBOSE
        if (block->boundary != NULL) {
          LOGDEBUG("  : Boundary     : %s", block->boundary);
        }
        if (block->terminating_boundary != NULL)
          LOGDEBUG("  : Term Boundary: %s", block->terminating_boundary);
        LOGDEBUG("  : Encoding     : %d", block->encoding);
        LOGDEBUG("  : Media Type   : %d", block->media_type);
        LOGDEBUG("  : Media Subtype: %d", block->media_subtype);
        LOGDEBUG("  : Headers:");
        node_header = c_nt_first(block->headers, &c2);
        while(node_header != NULL) {
          struct _ds_header_field *header = (struct _ds_header_field *) node_header->ptr;
          LOGDEBUG("    %-32s  %s", header->heading, header->data);
          node_header = c_nt_next(block->headers, &c2);
        }
#endif

        if (block->encoding == EN_BASE64 || block->encoding == EN_QUOTED_PRINTABLE) { 
          LOGDEBUG("decoding message block from encoding type %d", block->encoding);
          body = _ds_decode_block(block);  

          /* 7bit is safest */
          block->encoding = EN_7BIT;

          node_header = c_nt_first(block->headers, &c2);
          while(node_header != NULL) {
            struct _ds_header_field *header = (struct _ds_header_field *) node_header->ptr;
            if (!strcasecmp(header->heading, "Content-Transfer-Encoding")) {
              free(header->data);
              header->data = strdup("7bit");
            }
            node_header = c_nt_next(block->headers, &c2);
          }

          buffer_destroy(block->body);
          block->body = buffer_create(body);
          free(body);

          body = block->body->data;

        } else
          body = block->body->data;

        /* Look for signature */
        if (body != NULL) {
          signature_begin = strstr(body, SIGNATURE_BEGIN);

          if (signature_begin == NULL)
            signature_begin = strstr(body, LOOSE_SIGNATURE_BEGIN);
                                                                                                                                                                     
          if (signature_begin != NULL) {
            erase_begin = signature_begin;
            signature_begin = strstr(signature_begin, SIGNATURE_DELIMITER)+strlen(SIGNATURE_DELIMITER);
            signature_end = signature_begin;

            /* Find the signature's end character */
            while(signature_end != NULL && signature_end[0] != 0 && (isalnum((int) signature_end[0]) || signature_end[0]==32) )
              signature_end++;
                                                                                                                                                                     
            if (signature_end != NULL) {
              signature_length = signature_end - signature_begin;

              if (signature_length<128) {
                memcpy(signature_key, signature_begin, signature_length);
                signature_key[signature_length] = 0;
                strcpy(erase_begin, signature_end+1);
                have_signature = 1;
                LOGDEBUG("found signature '%s'", signature_key);
              }
            }
          }
        }
      }
#endif
      prev_node = node_nt;
      node_nt = c_nt_next(components->components, &c);
      i++;
    }

    if (have_signature==1) {
      if (_ds_get_signature(CTX, &SIG, signature_key)) {
        LOGDEBUG("signature retrieval for '%s' failed", signature_key);
        have_signature = 0;
      } else {
        CTX->signature = &SIG;
        _ds_delete_signature(CTX, signature_key);
      }

    }
  }

  if (!have_signature && mode != DSM_PROCESS) {
    CTX->flags  = CTX->flags ^ DSF_SIGNATURE;
    CTX->signature = NULL;
  }

  if (have_signature && mode != DSM_PROCESS) 
    result = dspam_process(CTX, NULL);
  else {
    CTX->signature = NULL;
    if (mode == DSM_ADDSPAM && !(CTX->flags & DSF_CORPUS))
      CTX->flags |= DSF_IGNOREHEADER;
 
    result = dspam_process(CTX, message->data);
  }

  result = CTX->result;

  /* Save the DSPAM signature */
  if (mode == DSM_PROCESS && CTX->signature != NULL) {
    int valid = 0;

    while(valid == 0) {
      _ds_create_signature_id(CTX, session, sizeof(session));
#ifndef SIGNATURE_ATTACHMENTS
      if (_ds_verify_signature(CTX, session))
#endif
        valid = 1;
    }

#ifdef SIGNATURE_ATTACHMENTS

    /* Embed the signature as an attachment; convert email if necessary */

    node_nt = c_nt_first(CTX->message->components, &c_nt);
    if (node_nt != NULL) {
      struct _ds_message_block *block = node_nt->ptr;
      if (block != NULL && block->headers != NULL) {
        long num_tokens, i;
        unsigned long long token;
        char boundary[128], t[128];
        buffer *s;
        int boundary_ok = 0;
#ifndef SIGNATURE_ATTACHMENTS
	char *encoded;
#endif

        boundary[0] = 0;
        s = buffer_create(NULL);

        if (s == NULL) {
          LOG(LOG_CRIT, ERROR_MEM_ALLOC);
          return EUNKNOWN;
        }

        num_tokens = CTX->signature->length / sizeof(unsigned long long);
        snprintf(t, sizeof(t), "%ld\n", num_tokens);
        buffer_cat(s, t);

        for(i=0;i<num_tokens;i++) {
          memcpy(&token, CTX->signature->data + (i*sizeof(unsigned long long)), sizeof(unsigned long long));
          snprintf(t, sizeof(t), "%016llx", token);

          buffer_cat(s, t);
          buffer_cat(s, "\n"); 
        }

        buffer_cat(s, "\n");

        /* Message is not multipart; convert into multipart:

           - Rewrite message header; add multipart boundary
           - Move body to new block with original headers

        */

        if (block->media_type != MT_MULTIPART) {
          struct _ds_header_field *header, *original_header = NULL, *original_encoding = NULL;
          struct _ds_message_block *newblock;
          struct nt_node *nt_header, *prev = NULL;
          struct nt_c c_header;

          boundary_ok = 1;
          snprintf(boundary, sizeof(boundary), "=ds_%s=", session);
 
          /* Rewrite message header */

          nt_header = c_nt_first(block->headers, &c_header);
          while(nt_header != NULL) {
            char typedata[512];
            header = (struct _ds_header_field *) nt_header->ptr;
            if (!strcasecmp(header->heading, "Content-Type")) {
              original_header = header;
              nt_header->ptr = (void *) malloc(sizeof(struct _ds_header_field));
              if (nt_header->ptr == NULL) {
                LOG(LOG_CRIT, ERROR_MEM_ALLOC);
                return EUNKNOWN;
              }

              /* Replace original header */

              header = (struct _ds_header_field *) nt_header->ptr;
              header->heading = strdup("Content-Type");
              snprintf(typedata, sizeof(typedata), "multipart/mixed; boundary=\"%s\"", boundary); 
              header->data = strdup(typedata);
              block->terminating_boundary = strdup(boundary); 
            } 

            /* Remove Encoding */
            if (!strcasecmp(header->heading, "Content-Transfer-Encoding")) {
              if (prev == NULL) 
                block->headers->first = nt_header->next;              
              else
                prev->next = nt_header->next;

              original_encoding = header;
            }

            prev = nt_header;
            nt_header = c_nt_next(block->headers, &c_header);
          }

          /* Add the header if there isn't one to replace */
          if (original_header == NULL) {
              char typedata[512];
              header = (struct _ds_header_field *) malloc(sizeof(struct _ds_header_field));
              if (header == NULL) {
                LOG(LOG_CRIT, ERROR_MEM_ALLOC);
                return EUNKNOWN;
              }
              header->heading = strdup("Content-Type");
              snprintf(typedata, sizeof(typedata), "multipart/mixed; boundary=\"%s\"", boundary);
              header->data = strdup(typedata);
              nt_add(block->headers, (void *) header);
              block->terminating_boundary = strdup(boundary);
          }

          /* Create new message block */ 
          newblock = (struct _ds_message_block *) malloc(sizeof(struct _ds_message_block));
          if (newblock == NULL) {
            LOG(LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          } 

          newblock->headers = nt_create(NT_PTR);
          if (newblock->headers == NULL) {
            free(newblock);
            LOG(LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }
 
          /* Move block type info to new block */

          newblock->terminating_boundary = strdup(boundary);
          newblock->encoding = block->encoding;
          newblock->original_encoding = block->original_encoding;
          newblock->media_type = block->media_type;
          newblock->media_subtype = block->media_subtype;
          block->encoding = EN_UNKNOWN;
          block->original_encoding = EN_UNKNOWN;
          block->media_type = MT_MULTIPART;
          block->media_subtype = MST_MIXED; 
          newblock->body = block->body;
          block->body = buffer_create(NULL);

          if (original_header != NULL)
            nt_add(newblock->headers, original_header);
          if (original_encoding != NULL)
          nt_add(newblock->headers, original_encoding);
 
          /* Add the new block */

          nt_add(CTX->message->components, (void *) newblock);
        }

        /* Message is multipart/alternative; convert into multipart/mixed

          ALTERNATIVE 

          - Move multipart/alternative heading to new multipart block in message
          - Rewrite message header as multipart/mixed
          - Create new block with message body

        */

        else if (block->media_subtype == MST_ALTERNATIVE) {
          struct nt_node *new_node = malloc(sizeof(struct nt_node));
          struct _ds_header_field *header, *original_header = NULL, *original_encoding = NULL;
          struct _ds_message_block *newblock;
          struct nt_node *nt_header, *prev = NULL;
          struct nt_c c_header;
          char old_boundary[sizeof(boundary)];

          boundary_ok = 1;
          snprintf(boundary, sizeof(boundary), "=ds_%s=", session);
          strlcpy(old_boundary, block->terminating_boundary, sizeof(old_boundary));

          /* Rewrite message header */
                                                                                                                                                                  
          nt_header = c_nt_first(block->headers, &c_header);
          while(nt_header != NULL) {
            char typedata[512];
            header = (struct _ds_header_field *) nt_header->ptr;
            if (!strcasecmp(header->heading, "Content-Type")) {
              original_header = header;
              nt_header->ptr = (void *) malloc(sizeof(struct _ds_header_field));
              if (nt_header->ptr == NULL) {
                LOG(LOG_CRIT, ERROR_MEM_ALLOC);
                return EUNKNOWN;
              }
                                                                                                                                                                  
              /* Replace original header */
                                                                                                                                                                  
              header = (struct _ds_header_field *) nt_header->ptr;
              header->heading = strdup("Content-Type");
              snprintf(typedata, sizeof(typedata), "multipart/mixed; boundary=\"%s\"", boundary);
              header->data = strdup(typedata);
              block->terminating_boundary = strdup(boundary);
            }

            /* Remove Encoding */
            if (!strcasecmp(header->heading, "Content-Transfer-Encoding")) {
              if (prev == NULL)
                block->headers->first = nt_header->next;
              else
                prev->next = nt_header->next;
                                                                                                                                                                  
              original_encoding = header;
            }
                                                                                                                                                                  
            prev = nt_header;
            nt_header = c_nt_next(block->headers, &c_header);
          }

          /* Create new message block */
          newblock = (struct _ds_message_block *) malloc(sizeof(struct _ds_message_block));
          if (newblock == NULL) {
            LOG(LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }
                                                                                                                                                                  
          newblock->headers = nt_create(NT_PTR);
          if (newblock->headers == NULL) {
            free(newblock);
            LOG(LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }

          if (new_node == NULL) {
            LOG(LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }

          /* Move block type info to new block */
          newblock->terminating_boundary = strdup(old_boundary);
          newblock->encoding = block->encoding;
          newblock->original_encoding = block->original_encoding;
          newblock->media_type = block->media_type;
          newblock->media_subtype = block->media_subtype;
          block->encoding = EN_UNKNOWN;
          block->original_encoding = EN_UNKNOWN;
          block->media_type = MT_MULTIPART;
          block->media_subtype = MST_ALTERNATIVE;
          newblock->body = block->body;
          block->body = buffer_create(NULL);
                                                                                                                                                                  
          if (original_header != NULL)
            nt_add(newblock->headers, original_header);
          if (original_encoding != NULL)
            nt_add(newblock->headers, original_encoding);
                                                                                                                                                                  
          /* Insert the new block before the alternative blocks */
          new_node->ptr = (void *) newblock;
          new_node->next = node_nt->next;
          node_nt->next = new_node;
 
          /* Add a new terminating block */
          /* Create new message block */
          newblock = (struct _ds_message_block *) malloc(sizeof(struct _ds_message_block));
          if (newblock == NULL) {
            LOG(LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }

          newblock->headers = nt_create(NT_PTR);
          if (newblock->headers == NULL) {
            free(newblock);
            LOG(LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }

          newblock->terminating_boundary = strdup(boundary);
          newblock->encoding = EN_UNKNOWN;
          newblock->original_encoding = EN_UNKNOWN;
          newblock->media_type = EN_UNKNOWN;
          newblock->media_subtype = EN_UNKNOWN;
          newblock->body = buffer_create(NULL);
          nt_add(CTX->message->components, (void *) newblock); 

        }

        /* 
           Message is now in an acceptable multipart format;
           insert signature attachment 
        */

        /* Strip off the -- on the last terminating boundary */
        if (! boundary_ok) {
          char term[sizeof(boundary)];
          strlcpy(boundary, block->boundary, sizeof(boundary));
          snprintf(term, sizeof(term), "%s--", boundary);

          while(node_nt != NULL && term[0] != 0) {
             block = (struct _ds_message_block *) node_nt->ptr;
            if (block->terminating_boundary != NULL && !strcmp(block->terminating_boundary, term)) {
              free(block->terminating_boundary);
              block->terminating_boundary = strdup(boundary);
              term[0] = 0;
            }
            node_nt = c_nt_next(CTX->message->components, &c_nt);
          }
          node_nt = c_nt_first(CTX->message->components, &c_nt);
          block = (struct _ds_message_block *) node_nt->ptr;
        }

        if (boundary[0] != 0) {
          struct _ds_message_block *newblock;
          struct _ds_header_field *field;
          char term[512];

         /* Create new message block */
          newblock = (struct _ds_message_block *) malloc(sizeof(struct _ds_message_block));
          if (newblock == NULL) {
            LOG(LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }

          newblock->headers = nt_create(NT_PTR);
          if (newblock->headers == NULL) {
            free(newblock);
            LOG(LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }
                                                                                                                                                                    
          /* Create new block information */
          snprintf(term, sizeof(term), "%s--", boundary);
                                                                                                                                                                    
          newblock->terminating_boundary = strdup(term);
          newblock->encoding = EN_BASE64;
          newblock->original_encoding = EN_BASE64;
          newblock->media_type = MT_APPLICATION;
          newblock->media_subtype = MST_DSPAM_SIGNATURE;
          newblock->body = s;

          field = (struct _ds_header_field *) malloc(sizeof(struct _ds_header_field));
          if (field == NULL) {
            _ds_destroy_block(newblock);
            LOG(LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }
          field->heading = strdup("Content-Type");
          field->data = strdup("application/dspam-signature; name=\"dspam.dat\"");
          nt_add(newblock->headers, field);

          field = (struct _ds_header_field *) malloc(sizeof(struct _ds_header_field));
          if (field == NULL) {
            _ds_destroy_block(newblock);
            LOG(LOG_CRIT, ERROR_MEM_ALLOC);
            return EUNKNOWN;
          }
          field->heading = strdup("Content-Transfer-Encoding");
          field->data = strdup("7bit");
          nt_add(newblock->headers, field);
                                                                                                                                                                    
          /* Add the new block */
                                                                                                                                                                    
          nt_add(CTX->message->components, (void *) newblock);

        }

        
      }
    }

#else
    _ds_set_signature(CTX, CTX->signature, session);
#endif
  }

  write_web_stats(username, &CTX->totals);

  LOGDEBUG("libdspam returned probability of %f, message result: %s", CTX->probability, (result != DSR_ISSPAM) ? "NOT SPAM" : "SPAM");

  /* FP and SM can return */
  if (CTX->message == NULL) {
    dspam_destroy(CTX);
    return result;
  }

  node_nt = c_nt_first(CTX->message->components, &c_nt);
  if (node_nt != NULL) {
    struct _ds_message_block *block = node_nt->ptr;
    if (block != NULL && block->headers != NULL) {
      char x[16];

      struct _ds_header_field *head = malloc(sizeof(struct _ds_header_field));
      if (head != NULL) {
        head->heading = strdup("X-DSPAM-Result");
        head->data = strdup((result == DSR_ISSPAM) ? "Spam" : "Innocent");
#ifdef VERBOSE
        LOGDEBUG("appending header %s: %s", head->heading, head->data)
#endif
        nt_add(block->headers, (void *) head);
      } else
        LOG(LOG_CRIT, ERROR_MEM_ALLOC);

      head = malloc(sizeof(struct _ds_header_field));
      if (head != NULL) {
        head->heading = strdup("X-DSPAM-Probability");
        snprintf(x, sizeof(x), "%f", CTX->probability);
        head->data = strdup(x);
#ifdef VERBOSE
        LOGDEBUG("appending header %s: %s", head->heading, head->data)
#endif
        nt_add(block->headers, (void *) head);
      } else
        LOG(LOG_CRIT, ERROR_MEM_ALLOC);
    }
  }

#ifndef SIGNATURE_ATTACHMENTS
  /* Embed the signature into all text segments */
  i = 0;
  while(node_nt != NULL) {
    struct _ds_message_block *block = node_nt->ptr;
    char *body_close = NULL,*dup = NULL;
    if (block != NULL && (block->media_type == MT_TEXT || (block->boundary == NULL && i == 0 && block->media_type != MT_MULTIPART))) {
      int unclosed_html = 0, is_attachment = 0;
      struct _ds_header_field *field;
      struct nt_node *node_hnt;
      struct nt_c c_hnt;

      node_hnt = c_nt_first(block->headers, &c_hnt);
      while(node_hnt != NULL) {
        field = (struct _ds_header_field *) node_hnt->ptr;
        if (field != NULL && field->heading != NULL && field->data != NULL) 
          if (!strncasecmp(field->heading, "Content-Disposition", 19))
            if (!strncasecmp(field->data, "attachment", 10))
              is_attachment = 1;
        node_hnt = c_nt_next(block->headers, &c_hnt);
      }

      if (is_attachment) {
        node_nt = c_nt_next(CTX->message->components, &c_nt);
        i++;
        continue;
      }
      
      if (block->body != NULL && block->body->data != NULL && block->media_subtype == MST_HTML)
        body_close = strstr(block->body->data, "</body");
      if (body_close == NULL && block->body != NULL && block->body->data != NULL && block->media_subtype == MST_HTML) 
        body_close = strstr(block->body->data, "</BODY");
      if (body_close == NULL && block->body != NULL && block->body->data != NULL && block->media_subtype == MST_HTML)
        body_close = strstr(block->body->data, "</Body");

      if (body_close != NULL) {

        dup = strdup(body_close);
        block->body->used -= strlen(dup);
        body_close[0] = 0;
      } 

      if (block->body->data != NULL && ( (strstr(block->body->data, "</html") || strstr(block->body->data, "</HTML")) && (!strstr(block->body->data, "</html>") && !strstr(block->body->data, "</HTML>")) ))
        unclosed_html = 1;

      if (!dup && unclosed_html)
        buffer_cat(block->body, ">");

      buffer_cat(block->body, "\n");
      buffer_cat(block->body, SIGNATURE_BEGIN);
      buffer_cat(block->body, session);
      buffer_cat(block->body, SIGNATURE_END);
      buffer_cat(block->body, "\n\n");
 
      if (dup) {
        buffer_cat(block->body, dup);
        free(dup);
        if (unclosed_html) 
          buffer_cat(block->body, ">");
      }
    }
    node_nt = c_nt_next(CTX->message->components, &c_nt);
    i++;
  }

#endif

  /* reconstruct message from components */

  copyback = _ds_assemble_message(CTX->message);
  buffer_clear(message);
  buffer_cat(message, copyback);
  free(copyback);

  if (CTX->message != NULL)
    _ds_destroy_message(CTX->message);
  dspam_destroy(CTX);

  return result;
}
 
int deliver_message(const char *message, const char *mailer_args, const char *username, int mode) {
#ifdef DELIVER_TO_STDOUT
  fputs(message, stdout);
#else
  char args[1024];
  char *margs, *mmargs, *arg;
  FILE *file;

  args[0] = 0;
  margs = strdup(mailer_args);
  mmargs = margs;
  arg = strsep(&margs, " ");
  while(arg != NULL) {
    if (!strcmp(arg, "$u") || !strcmp(arg, "\\$u") || !strcmp(arg, "%u"))
      strlcat(args, username, sizeof(args));
    else
      strlcat(args, arg, sizeof(args));
    strlcat(args, " ", sizeof(args));
    arg = strsep(&margs, " ");
  }
  free(mmargs);
  
  LOGDEBUG("Opening pipe to LDA: %s", args);
  file = popen(args, "w");

  if (file == NULL) { 
    file_error(ERROR_FILE_OPEN, args, strerror(errno));
    return EFILE;
  }

  fputs(message, file);
  fclose(file);
#endif

  return 0;
}

int quarantine_message(const char *message, const char *username) {
  char filename[MAX_FILENAME_LENGTH];
  FILE *file;

  if (quarantine_lock(username)) 
    return -1;

  strcpy(filename, _ds_userdir_path(username, "mbox"));
  file = fopen(filename, "a");
  if (file == NULL) {
    file_error(ERROR_FILE_WRITE, filename, strerror(errno));
    quarantine_unlock(username);
    return EFILE;
  }

  /* From header */
  if (strncmp(message, "From ", 5)) {
    char head[128];
    time_t tm = time(NULL);

    snprintf(head, sizeof(head), "From %s %s", username, ctime(&tm));
    fputs(head, file);
  }

  fputs(message, file);
  fputs("\n\n", file);

  fclose(file);

  quarantine_unlock(username); 
  return 0;
} 

int write_web_stats(const char *username, struct _ds_spam_totals *totals) {
  char filename[MAX_FILENAME_LENGTH];
  FILE *file;

  if (totals == NULL) {
    LOGDEBUG("totals are null\n");
    return EINVAL;
  }

  strcpy(filename, _ds_userdir_path(username, "stats"));
  file = fopen(filename, "w");
  if (file == NULL) {
    file_error(ERROR_FILE_WRITE, filename, strerror(errno));
    return EFILE;
  }

  fprintf(file, "%ld,%ld,%ld,%ld\n", totals->total_spam, totals->total_innocent, totals->spam_misses, totals->false_positives);
  fclose(file);
  return 0;
}

int quarantine_lock(const char *username) {
  char filename[MAX_FILENAME_LENGTH];
  int retry = 0, pid;
  char scratch[1024];
  char myhostname[1024];
  char *hostname, *spid;
  FILE *file;
  struct stat s;

  gethostname(myhostname, sizeof(myhostname));

  strcpy(filename, _ds_userdir_path(username, "qlock"));

  LOGDEBUG("acquiring quarantine lock %s", filename);

  file = fopen(filename, "r");

  while(file != NULL) {
    if(fgets(scratch, sizeof(scratch), file)!=NULL) {
      LOGDEBUG("waiting for quarantine lock");
      hostname = strtok(scratch, ":");
      spid = strtok(NULL, ":");
      fclose(file);
      if (hostname != NULL && !strcmp(hostname, myhostname) && spid != NULL) {
        LOGDEBUG("hostname matches %s.  checking pid", hostname);
        pid = atoi(spid);
        if (kill(pid, 0)) {
          retry = 0;
          unlink(filename);
        } else {
          retry++;
        }
      } else {
        retry++;
        if (!stat(filename, &s)) {
          if (time(NULL) - s.st_mtime > 60*20) {
            LOGDEBUG("removing stale quarantine lock %s", filename);
            retry = 0;
            unlink(filename);
          }
        } else {
          retry = 0;
          unlink(filename);
        }
      }

    } else {
      retry = 0;
      fclose(file);
      unlink(filename);
    }

    if (retry>=LOCK_TIMEOUT) {
      LOG(LOG_NOTICE, "quarantine lock timeout");
      return ELOCK;
    }

    if (retry) 
      sleep(1);

    file = fopen(filename, "r");
  }

  file = fopen(filename, "w");
  if (file != NULL) {
    fprintf(file, "%s:%ld\n", myhostname, (long) getpid());
    fclose(file);
  } else {
    LOG(LOG_WARNING, "quarantine lock failed: %s: could not create file", strerror(errno));
    return EFILE;
  }

  /* Verify the quarantine lock */
  file = fopen(filename, "r");
  if (file == NULL) {
    LOG(LOG_WARNING, "quarantine lock failed: %s: could not create file", strerror(errno));
    return EFILE;
  }

  if (fgets(scratch, sizeof(scratch), file)==NULL) {
    LOG(LOG_WARNING, "quarantine lock failed: lock pid does not match mine");
    fclose(file);
    return quarantine_lock(username);
  }

  fclose(file);

  hostname = strtok(scratch, ":");
  spid = strtok(NULL, ":");
  if (hostname == NULL || spid == NULL || strcmp(hostname, myhostname) || atoi(spid)!=getpid()) {
    LOGDEBUG("quarantine lock failed.  retrying.");
    return quarantine_lock(username);
  }
  
  return 0;
}

int quarantine_unlock(const char *username) {
  char filename[MAX_FILENAME_LENGTH];

  strcpy(filename, _ds_userdir_path(username, "qlock"));

  LOGDEBUG("freeing quarantine lock for '%s': %s", username, filename);
  return unlink(filename);
}

