/*
 DSPAM
 COPYRIGHT (C) 2003 NETWORK DWEEBS CORPORATION

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <db.h>

#ifdef TIME_WITH_SYS_TIME
#   include <sys/time.h>
#   include <time.h>
#else
#   ifdef HAVE_SYS_TIME_H
#       include <sys/time.h>
#   else
#       include <time.h>
#   endif
#endif

#if DB_VERSION_MINOR == 0
#	define DBOPEN(A, B, C, D, E, F, G) open(A, C, D, E, F, G)
#else
#	define DBOPEN(A, B, C, D, E, F, G) open(A, B, C, D, E, F, G)
#endif

#include "storage_driver.h"
#include "libdb4_drv.h"
#include "libdspam.h"
#include "config.h"
#include "error.h"
#include "language.h"
#include "util.h"
#include "lht.h"

int _libdb4_drv_get_spamtotals(DSPAM_CTX *CTX) {
  DBT key, data;
  char hashkey[32];
  int ret;
  struct _libdb4_drv_storage *s = (struct _libdb4_drv_storage *) CTX->storage;

  if (s->db == NULL) 
    return EINVAL;
    
  memset(&key, 0, sizeof(DBT));
  memset(&data, 0, sizeof(DBT));

  strcpy(hashkey, "_TOTALS");
  key.data = hashkey;
  key.size = strlen(hashkey);

  ret = (*s->db->get)(s->db, NULL, &key, &data, 0);
  if (ret) {
    LOGDEBUG("_ds_get_totals: db->get failed: %s", db_strerror(ret));
    return EFAILURE;
  }

  if (data.size == sizeof(struct _ds_spam_totals)) 
    memcpy(&CTX->totals, data.data, data.size);
  else {
    LOG(LOG_WARNING, "_ds_get_totals: record size (%d) doesn't match sizeof(struct _ds_spam_totals) (%d)", data.size, sizeof(struct _ds_spam_totals));
    return EUNKNOWN;
  }
  
  return 0;
}

int _libdb4_drv_set_spamtotals(DSPAM_CTX *CTX) {
  DBT key, data;
  char hashkey[32];
  int ret;
  struct _libdb4_drv_storage *s = (struct _libdb4_drv_storage *) CTX->storage;

  if (s->db == NULL)
    return EINVAL;

  if (CTX->flags & DSF_CLASSIFY) {
    _libdb4_drv_get_spamtotals(CTX); /* undo changes to in memory totals */
    return 0;
  }

  memset(&key, 0, sizeof(DBT));
  memset(&data, 0, sizeof(DBT));

  strcpy(hashkey, "_TOTALS");
  key.data = hashkey;
  key.size = strlen(hashkey);
  data.data = &CTX->totals;
  data.size = sizeof(struct _ds_spam_totals);

  ret = (*s->db->put)(s->db, NULL, &key, &data, 0);
  if (ret) {
    LOG(LOG_WARNING, "_ds_set_totals: db->put failed: %s", db_strerror(ret));
    return EFAILURE;
  }

  return 0;
}

int _ds_getall_spamrecords(DSPAM_CTX *CTX, struct lht *freq) {
  struct lht_node * node_lht;
  struct lht_c c_lht;
  struct _ds_spam_stat stat;
  int ret = 0, x = 0;

  if (freq == NULL || CTX == NULL)
    return EINVAL;

  node_lht = c_lht_first(freq, &c_lht);
  while(node_lht != NULL) {
    node_lht->s.spam_hits = 0;
    node_lht->s.innocent_hits = 0;
    x = _ds_get_spamrecord(CTX, node_lht->key, &stat);
    if (!x) {
      lht_setspamstat(freq, node_lht->key, &stat);
      if (node_lht->s.spam_hits > CTX->totals.total_spam)
        node_lht->s.spam_hits = CTX->totals.total_spam;
      if (node_lht->s.innocent_hits > CTX->totals.total_innocent)
        node_lht->s.innocent_hits = CTX->totals.total_innocent;
    } else if (x != EFAILURE) {
      ret = x;
    }

    node_lht = c_lht_next(freq, &c_lht);
  }

  return ret;
}

int _ds_setall_spamrecords(DSPAM_CTX *CTX, struct lht *freq) {
  struct lht_node * node_lht;
  struct lht_c c_lht;
  int ret = EUNKNOWN;
                                                                                                                                                                                
  if (freq == NULL || CTX == NULL)
    return EINVAL;
                                                                                                                                                                                
  node_lht = c_lht_first(freq, &c_lht);
  while(node_lht != NULL) {
    if (!_ds_set_spamrecord(CTX, node_lht->key, &node_lht->s))
      ret = 0;
    node_lht = c_lht_next(freq, &c_lht);
  }
                                                                                                                                                                                
  return ret;
}

int _ds_get_spamrecord(DSPAM_CTX *CTX, unsigned long long token, struct _ds_spam_stat *stat) {
  struct _libdb4_drv_spam_record record;
  unsigned long long value = token;
  DBT key, data;
  int ret;
  struct _libdb4_drv_storage *s = (struct _libdb4_drv_storage *) CTX->storage;

  if (s->db == NULL)
    return EINVAL;

  memset(&key, 0, sizeof(DBT));
  memset(&data, 0, sizeof(DBT));

  key.data = &value;
  key.size = sizeof(unsigned long long);

  ret = (*s->db->get)(s->db, NULL, &key, &data, 0);

  if (ret) 
    return EFAILURE;

  if (data.size == sizeof(struct _libdb4_drv_spam_record)) {
    memcpy(&record, data.data, data.size);
    stat->spam_hits	= record.spam_hits;
    stat->innocent_hits	= record.innocent_hits;
  }
  else {
    LOG(LOG_WARNING, "_ds_get_spamrecord: record size (%d) doesn't match sizeof(struct _libdb4_drv_spam_record) (%d)", data.size, sizeof(struct _libdb4_drv_spam_record));
    return EUNKNOWN;
  }

  return 0;
}

int _ds_set_spamrecord(DSPAM_CTX *CTX, unsigned long long token, struct _ds_spam_stat *stat) {
  struct _libdb4_drv_spam_record record;
  unsigned long long value = token;
  DBT key, data;
  int ret;
  struct _libdb4_drv_storage *s = (struct _libdb4_drv_storage *) CTX->storage;

  if (s->db == NULL)
    return EINVAL;

  if (CTX->flags & DSF_CLASSIFY) 
    return 0;

  memset(&key, 0, sizeof(DBT));
  memset(&data, 0, sizeof(DBT));

  if (CTX->_process_start == 0)
    CTX->_process_start = time(NULL);

  record.spam_hits      = (stat->spam_hits>0) ? stat->spam_hits : 0;
  record.innocent_hits  = (stat->innocent_hits>0) ? stat->innocent_hits : 0;
  record.last_hit       = CTX->_process_start;

  key.data = &value;
  key.size = sizeof(unsigned long long);
  data.data = &record;
  data.size = sizeof(struct _libdb4_drv_spam_record);

  ret = (*s->db->put)(s->db, NULL, &key, &data, 0);

  if (ret) {
    LOG(LOG_WARNING, "_ds_set_spamrecord: db->put failed: %s", db_strerror(ret));
    return EFAILURE;
  }

  return 0;
}

int _ds_init_storage(DSPAM_CTX *CTX) {
  struct _libdb4_drv_storage *s;
  int dbflag = DB_CREATE;

  /* don't init if we're already initted */
  if (CTX->storage != NULL) 
    return EINVAL;

  if (CTX->flags & DSF_CLASSIFY)
    dbflag = DB_RDONLY;

  s = malloc(sizeof(struct _libdb4_drv_storage));
  if (s == NULL) {
    LOG(LOG_CRIT, ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }

  s->db = NULL;
  s->sig = NULL;

  if (CTX->username != NULL) {

    if (_libdb4_drv_context_lock(CTX))
      return ELOCK;

    if (CTX->group == NULL) { 
      strcpy(s->dictionary, _ds_userdir_path(CTX->username, "dict"));
      strcpy(s->signature, _ds_userdir_path(CTX->username, "sig"));
    } else {
      strcpy(s->dictionary, _ds_userdir_path(CTX->group, "dict"));
      strcpy(s->signature, _ds_userdir_path(CTX->group, "sig"));
    }

    _libdb4_drv_prepare_path_for(s->dictionary);
  
    if ((CTX->result = db_create(&s->db, NULL, 0))!=0) {
      LOG(LOG_WARNING, "db_create failed: %s", db_strerror(CTX->result));
      s->db = NULL;
      return EFAILURE;
    }
                                                                                                                                                                            
    if ((s->db->set_cachesize(s->db, 0, 1024*1024*DB_CACHESIZE_MB, 1))!=0) {
      LOG(LOG_WARNING, "db->set_cachesize failed: %s", db_strerror(CTX->result));
      s->db = NULL;
      return EFAILURE;
    }
                                                                                                                                                                            
    if ((CTX->result = s->db->DBOPEN(s->db, NULL, s->dictionary, NULL, DB_BTREE, dbflag, 0))!=0) {
      LOG(LOG_WARNING, "db->open failed: %s: %s", s->dictionary, db_strerror(CTX->result));
      s->db = NULL;
      return EFILE;
    }
  
    if ((CTX->result = db_create(&s->sig, NULL, 0))!=0) {
      LOG(LOG_WARNING, "db_create failed: %s", db_strerror(CTX->result));
      s->sig = NULL;
      _ds_shutdown_storage(CTX);
      return EFAILURE;
    }
                                                                                                                                                                            
    if ((s->sig->set_cachesize(s->sig, 0, 1024*1024*DB_CACHESIZE_MB, 1))!=0) {
      LOG(LOG_WARNING, "db->set_cachesize failed: %s", db_strerror(CTX->result));
      s->sig = NULL;
      _ds_shutdown_storage(CTX);
      return EFAILURE;
    }
                                                                                                                                                                            
    if ((CTX->result = s->sig->DBOPEN(s->sig, NULL, s->signature, NULL, DB_BTREE, DB_CREATE, 0))!=0) {
      LOG(LOG_WARNING, "db->open failed: %s: %s", s->signature, db_strerror(CTX->result));
      s->sig = NULL;
      _ds_shutdown_storage(CTX);
      return EFILE;
    }
  } else {
    s->db = NULL;
    s->sig = NULL;
  }

  s->c = NULL;
  s->dir_handles = nt_create(NT_INDEX);
  CTX->storage = s;

  if (CTX->username != NULL) {
    if (_libdb4_drv_get_spamtotals(CTX)) {
      LOGDEBUG("unable to load totals.  using zero values.");
      memset(&CTX->totals, 0, sizeof(struct _ds_spam_totals));
    }
  } else {
      memset(&CTX->totals, 0, sizeof(struct _ds_spam_totals));
  }

  return 0;
}

int _ds_shutdown_storage(DSPAM_CTX *CTX) {
  struct _libdb4_drv_storage *s = (struct _libdb4_drv_storage *) CTX->storage;
  struct nt_node *node_nt;
  struct nt_c c_nt;
  int ret = 0;

  if (s == NULL)
    return EINVAL;

  if (CTX->username != NULL && ! (CTX->flags & DSF_CLASSIFY)) { 
    _libdb4_drv_set_spamtotals(CTX);
  }
  node_nt = c_nt_first(s->dir_handles, &c_nt);
  while(node_nt != NULL) {
    DIR *dir;
    dir = (DIR *) node_nt->ptr;
    closedir(dir);
    node_nt = c_nt_next(s->dir_handles, &c_nt);
  }

  nt_destroy(s->dir_handles);

  /* cursor from _ds_get_nexttoken */
  if (s->c)
    s->c->c_close(s->c);

  /* token database */
  if (s->db) {
    ret = s->db->close(s->db, 0);
    _libdb4_drv_context_unlock(CTX);
    if (ret)
      return EUNKNOWN;
  }

  /* signature database */
  if (s->sig) 
    ret = s->sig->close(s->sig, 0);

 if (!ret) {
    free(CTX->storage);
    CTX->storage = NULL;
  }

  return (!ret) ? 0 : EUNKNOWN;
}

int _ds_create_signature_id(DSPAM_CTX *CTX, char *buf, size_t len) {
  char session[64];
  char digit[6];
  int pid, j;

  pid = getpid();
  snprintf(session, sizeof(session), "%8lx%d", time(NULL), pid);

  for(j=0;j<2;j++) {
    snprintf(digit, 6, "%d", rand());
    strlcat(session, digit, 64);
  }

  strlcpy(buf, session, len);
  return 0;
}

int _ds_get_signature(DSPAM_CTX *CTX, struct _ds_spam_signature *SIG, const char *signature) {
  struct _libdb4_drv_storage *s = (struct _libdb4_drv_storage *) CTX->storage;
  DBT key, data;
  char *hashkey;
  void *mem;
  int ret;

  hashkey = malloc(strlen(signature)+1);

  memcpy(hashkey, signature, strlen(signature)+1);
 
  memset(&key, 0, sizeof(DBT));
  memset(&data, 0, sizeof(DBT));

  key.data = hashkey;
  key.size = strlen(hashkey);

  ret = (*s->sig->get)(s->sig, NULL, &key, &data, 0);

  if (ret) {
    free(hashkey);
    return EFAILURE;
  }

  mem = (void *) calloc(1, data.size-sizeof(time_t));
  if (mem == NULL) {
    free(hashkey);
    report_error(ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }
  memcpy(mem, (char *)data.data+sizeof(time_t), data.size-sizeof(time_t));
  SIG->data = mem;
  SIG->length = data.size-sizeof(time_t);
 
  free(hashkey);
  return 0;
}

int _ds_set_signature(DSPAM_CTX *CTX, struct _ds_spam_signature *SIG, const char *signature) {
  struct _libdb4_drv_storage *s = (struct _libdb4_drv_storage *) CTX->storage;
  DBT key, data;
  char *hashkey;
  void *mem;
  int ret;
  time_t t = time(NULL);

  hashkey = malloc(strlen(signature)+1);

  memcpy(hashkey, signature, strlen(signature)+1);
  mem = malloc(SIG->length+sizeof(time_t));
  if (mem == NULL) {
    free(hashkey);
    return EUNKNOWN;
  }

  memcpy(mem, &t, sizeof(time_t));
  memcpy((char *)mem+sizeof(time_t), SIG->data, SIG->length);

  memset(&key, 0, sizeof(DBT));
  memset(&data, 0, sizeof(DBT));

  key.data = hashkey;
  key.size = strlen(hashkey);
  data.data = mem;
  data.size = SIG->length+sizeof(time_t);;

  ret = (*s->sig->put)(s->sig, NULL, &key, &data, 0);

  if (ret) {
    free(hashkey);
    LOG(LOG_WARNING, "set_signature: sig->put failed: %s", db_strerror(ret));
    return EFAILURE;
  }

  free(hashkey);
  return 0;
}

int _ds_delete_signature(DSPAM_CTX *CTX, const char *signature) {
  struct _libdb4_drv_storage *s = (struct _libdb4_drv_storage *) CTX->storage;
  DBT key, data;
  char *hashkey;
  int ret;

  hashkey = malloc(strlen(signature)+1);

  LOGDEBUG("deleting signature %s", signature);
  memcpy(hashkey, signature, strlen(signature)+1);

  memset(&key, 0, sizeof(DBT));
  memset(&data, 0, sizeof(DBT));

  key.data = hashkey;
  key.size = strlen(hashkey);

  ret = (*s->sig->del)(s->sig, NULL, &key, 0);

  if (ret) {
    free(hashkey);
    LOGDEBUG("delete_signature: sig->del failed: %s", db_strerror(ret));
    return EFAILURE;
  }

  free(hashkey);
  return 0;
}

int _ds_verify_signature(DSPAM_CTX *CTX, const char *signature) {
  struct _libdb4_drv_storage *s = (struct _libdb4_drv_storage *) CTX->storage;
  DBT key, data;
  char *hashkey;

  if (s->sig == NULL) 
    return 0;

  hashkey = malloc(strlen(signature)+1);

  memcpy(hashkey, signature, strlen(signature)+1);

  memset(&key, 0, sizeof(DBT));
  memset(&data, 0, sizeof(DBT));

  key.data = hashkey;
  key.size = strlen(hashkey);

  free(hashkey);
  return((*s->sig->get)(s->sig, NULL, &key, &data, 0));
}

int _libdb4_drv_prepare_path_for(const char *filename) {
#ifdef LARGESCALE
  char path[MAX_FILENAME_LENGTH];
  char *dir, *file = strdup(filename);
  struct stat s;
                                                                                                                                                                          
  path[0] = 0;
  dir = strsep(&file, "/");
  while(dir != NULL) {
    strlcat(path, dir, sizeof(path));
    strlcat(path, "/", sizeof(path));
    dir = strsep(&file, "/");
                                                                                                                                                                          
    if (dir != NULL && stat(path, &s)) {
      int x;
      LOGDEBUG("creating directory '%s'", path);
      x = mkdir(path, 0770);
      if (x) {
        file_error(ERROR_DIR_CREATE, path, strerror(errno));
        return EFILE;
      }
    }
  }
  free(file);
#endif
  return 0;
}

int _libdb4_drv_context_lock(DSPAM_CTX *CTX) {
  char filename[MAX_FILENAME_LENGTH];
  int retry = 0, pid;
  char scratch[1024];
  char myhostname[1024];
  char *hostname, *spid;
  FILE *file;
  struct stat s;

  gethostname(myhostname, sizeof(myhostname));
  if (CTX->locked)
    return 0;

  if (CTX->group == NULL) {
    strcpy(filename, _ds_userdir_path(CTX->username, "lock"));
  } else {
    strcpy(filename, _ds_userdir_path(CTX->group, "lock"));
  }

  LOGDEBUG("acquiring lock %s", filename);

  file = fopen(filename, "r");

  while(file != NULL) {
    if(fgets(scratch, sizeof(scratch), file)!=NULL) {
      LOGDEBUG("waiting for lock");
      hostname = strtok(scratch, ":");
      spid = strtok(NULL, ":");
      fclose(file);
      if (hostname != NULL && !strcmp(hostname, myhostname) && spid != NULL) {
        LOGDEBUG("hostname matches %s.  checking pid", hostname);
        pid = atoi(spid);
        if (kill(pid, 0)) {
          retry = 0;
          unlink(filename);
        } else {
          retry++;
        }
      } else {
        retry++;
        if (!stat(filename, &s)) {
          if (time(NULL) - s.st_mtime > 60*20) {
            LOGDEBUG("removing stale lock %s", filename);
            retry = 0;
            unlink(filename);
          }
        } else {
          retry = 0;
          unlink(filename);
        }
      }

    } else {
      retry = 0;
      fclose(file);
      unlink(filename);
    }

    if (retry>=LOCK_TIMEOUT) {
      LOG(LOG_NOTICE, "lock timeout");
      return ELOCK;
    }

    if (retry) 
      sleep(1);

    file = fopen(filename, "r");
  }

  file = fopen(filename, "w");
  if (file != NULL) {
    fprintf(file, "%s:%ld\n", myhostname, (long) getpid());
    fclose(file);
  } else {
    LOG(LOG_WARNING, "lock failed: %s: could not create file", strerror(errno));
    return EFILE;
  }

  /* Verify the lock */
  file = fopen(filename, "r");
  if (file == NULL) {
    LOG(LOG_WARNING, "lock failed: %s: could not create file", strerror(errno));
    return EFILE;
  }

  if (fgets(scratch, sizeof(scratch), file)==NULL) {
    LOG(LOG_WARNING, "lock failed: lock pid does not match mine");
    fclose(file);
    return _libdb4_drv_context_lock(CTX);
  }

  fclose(file);

  hostname = strtok(scratch, ":");
  spid = strtok(NULL, ":");
  if (hostname == NULL || spid == NULL || strcmp(hostname, myhostname) || atoi(spid)!=getpid()) {
    LOGDEBUG("lock failed.  retrying.");
    return _libdb4_drv_context_lock(CTX);
  }
  
  CTX->locked = 1;
  return 0;
}

int _libdb4_drv_context_unlock(DSPAM_CTX *CTX) {
  char filename[MAX_FILENAME_LENGTH];

  if (! CTX->locked)
    return 0;
 
  if (CTX->group == NULL) {
    strcpy(filename, _ds_userdir_path(CTX->username, "lock"));
  } else {
    strcpy(filename, _ds_userdir_path(CTX->group, "lock"));
  }

  LOGDEBUG("freeing lock for '%s': %s", CTX->username, filename);
  CTX->locked = unlink(filename);
  return (CTX->locked) ? EFAILURE : 0;
}

char *_ds_get_nextuser(DSPAM_CTX *CTX) {
  static char user[MAX_FILENAME_LENGTH];
  static char path[MAX_FILENAME_LENGTH];
  struct _libdb4_drv_storage *s = (struct _libdb4_drv_storage *) CTX->storage;
  struct nt_node *node_nt, *prev;
  struct nt_c c_nt;
  char *x = NULL, *y;
  DIR *dir = NULL;

  struct dirent *entry;

  if (s->dir_handles->items == 0) {
    dir = opendir(USERDIR);
    if (dir == NULL) {
      LOG(LOG_WARNING, "unable to open directory '%s' for reading: %s", USERDIR, strerror(errno));
      return NULL; 
    }

    nt_add(s->dir_handles, (void *) dir);
    strlcpy(path, USERDIR, sizeof(path));
  } else {

    node_nt = c_nt_first(s->dir_handles, &c_nt);
    while(node_nt != NULL) {
      if (node_nt->next == NULL)
        dir = (DIR *) node_nt->ptr;
      node_nt = c_nt_next(s->dir_handles, &c_nt);
    }
  }

  while((entry = readdir(dir)) != NULL) {
    struct stat st;
    char filename[MAX_FILENAME_LENGTH];
    snprintf(filename, sizeof(filename), "%s/%s", path, entry->d_name);

    if (!strcmp(entry->d_name, ".") || !strcmp(entry->d_name, ".."))
      continue;

    if (stat(filename, &st))
      continue;

    /* push a new directory */
    if (st.st_mode & S_IFDIR) {
      dir = opendir(filename);
      if (dir == NULL)
        continue;
      strlcat(path, "/", sizeof(path));
      strlcat(path, entry->d_name, sizeof(path));
      nt_add(s->dir_handles, (void *) dir);
      return _ds_get_nextuser(CTX);
    } else if (!strncmp(entry->d_name+strlen(entry->d_name)-5, ".dict", 5)) {
      strlcpy(user, entry->d_name, sizeof(user));
      user[strlen(user)-5]=0;
      return user;
    }
  }

  /* pop current directory */
  y = strchr(path, '/');
  while(y != NULL) {
    x = y;
    y = strchr(x+1, '/'); 
  }
  if (x)
    x[0] = 0;

  /* pop directory handle from list */
  node_nt = c_nt_first(s->dir_handles, &c_nt);
  prev = NULL;
  while(node_nt != NULL) {
    if (node_nt->next == NULL) {
      dir = (DIR *) node_nt->ptr;
      closedir(dir);
      if (prev != NULL) 
        prev->next = NULL;
      else 
        s->dir_handles->first = NULL;
      free(node_nt);
      s->dir_handles->items--;
    }
    prev = node_nt;
    node_nt = c_nt_next(s->dir_handles, &c_nt);  
  }
  if (s->dir_handles->items>0)
    return _ds_get_nextuser(CTX);

  /* done */

  user[0] = 0;
  return NULL;
}

struct _ds_storage_record *_ds_get_nexttoken(DSPAM_CTX *CTX) {
  struct _ds_storage_record *sr;
  struct _libdb4_drv_storage *s = (struct _libdb4_drv_storage *) CTX->storage;
  DBT key, data;

  sr = malloc(sizeof(struct _ds_storage_record));
  if (sr == NULL) {
    LOG(LOG_CRIT, ERROR_MEM_ALLOC);
    return NULL;
  }

  if (s->c == NULL) 
    s->db->cursor(s->db, NULL, &s->c, 0);
                                                                                                                                                                           
  memset(&key, 0, sizeof(DBT));
  memset(&data, 0, sizeof(DBT));
                                                                                                                                                                           
  while(s->c->c_get(s->c, &key, &data, DB_NEXT)==0) {
    struct _libdb4_drv_spam_record record;
    char keyname[64];
                                                                                                                                                                           
    memcpy(&keyname, key.data, key.size);
    keyname[key.size] = 0;
    if (strcmp(keyname, "_TOTALS")) {
      snprintf(keyname, sizeof(keyname), "%llu", *(unsigned long long *)key.data);
      memcpy(&record, data.data, sizeof(struct _libdb4_drv_spam_record));
      sr->token = *(unsigned long long *)key.data;
      sr->spam_hits = record.spam_hits;
      sr->innocent_hits = record.innocent_hits;
      sr->last_hit = record.last_hit;
      return sr;
    }
  }
                                                                                                                                                                           
  s->c->c_close(s->c);
  s->c = NULL;
  return NULL;
}

struct _ds_storage_signature *_ds_get_nextsignature(DSPAM_CTX *CTX) {
  struct _ds_storage_signature *ss;
  struct _libdb4_drv_storage *s = (struct _libdb4_drv_storage *) CTX->storage;
  DBT key, data;

  ss = malloc(sizeof(struct _ds_storage_signature));
  if (ss == NULL) {
    LOG(LOG_CRIT, ERROR_MEM_ALLOC);
    return NULL;
  }

  if (s->c == NULL)
    s->sig->cursor(s->sig, NULL, &s->c, 0);

  memset(&key, 0, sizeof(DBT));
  memset(&data, 0, sizeof(DBT));

  while(s->c->c_get(s->c, &key, &data, DB_NEXT)==0) {
    ss->data = (void *) calloc(1, data.size-sizeof(time_t));
    if (ss->data == NULL) {
      LOG(LOG_CRIT, ERROR_MEM_ALLOC);
      free(ss);
      return NULL;
    }

    memcpy(ss->signature, key.data, key.size);
    ss->signature[key.size] = 0;
    ss->length = data.size-sizeof(time_t);
    memcpy(ss->data, (char *)data.data+sizeof(time_t), data.size-sizeof(time_t));
    memcpy(&ss->created_on, data.data, sizeof(time_t));
    return ss;
  }

  s->c->c_close(s->c);
  s->c = NULL;
  return NULL;
}

