/*
 DSPAM
 COPYRIGHT (C) 2003 NETWORK DWEEBS CORPORATION

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <db.h>

#ifdef TIME_WITH_SYS_TIME
#   include <sys/time.h>
#   include <time.h>
#else
#   ifdef HAVE_SYS_TIME_H
#       include <sys/time.h>
#   else
#       include <time.h>
#   endif
#endif

#include "libdspam_objects.h"
#include "libdspam.h"
#include "nodetree.h"
#include "config.h"
#include "base64.h"
#include "util.h"
#include "storage_driver.h"
#include "buffer.h"
#include "lht.h"
#include "tbt.h"
#include "error.h"
#include "decode.h"
#include "language.h"

/* dspam_init: initializes a DSPAM context 
   parameters: const char *dictionary	filename of dictionary
               int mode			processing mode (libdspam.h)
               int flags                processing flags (libdspam.h)

      returns: pointer to the newly allocated DSPAM context 
*/

DSPAM_CTX *dspam_init(const char *username, const char *group, int mode, int flags) {
  DSPAM_CTX *CTX;

  CTX = calloc(1, sizeof(DSPAM_CTX));
  if (CTX == NULL)
    return NULL;

  if (username != NULL && username[0] != 0)  
    CTX->username    = strdup(username);
  else
    CTX->username = NULL;

  if (group != NULL && group[0]!=0)
    CTX->group       = strdup(group);
  else
    CTX->group = NULL;
  CTX->probability = -1;
  CTX->mode        = mode;
  CTX->flags       = flags;
  CTX->message	   = NULL;

  if (!_ds_init_storage(CTX))
    return CTX;

  return NULL;
}


/* dspam_destroy: destroys a DSPAM context
      parameters: DSPAM_CTX *	pointer to context
         returns: 0 on successful completion
*/

int dspam_destroy(DSPAM_CTX *CTX) {
  if (CTX->storage != NULL)
    _ds_shutdown_storage(CTX);
  free(CTX->username);
  free(CTX->group);
  if (CTX->signature != NULL && CTX->mode == DSM_PROCESS)
    free(CTX->signature);

  free(CTX);
  return 0;
}

/* dspam_getresult: returns results of dspam_process() function
        parameters: DSPAM_CTX *	pointer to context
           returns: DSR_ISSPAM     - Is spam
                    DSR_ISINNOCENT - Is not spam 
                    <Other>        - Error code if dspam_process returned a
                                     non-zero value
*/

int dspam_getresult(DSPAM_CTX *CTX) {
  return CTX->result;
}

/* dspam_process: processes a message using the specified mode and flags 
      parameters: DSPAM_CTX *			pointer to context
                  const char *message		pointer to message text
         returns: 0 on successful completion
*/
int dspam_process(DSPAM_CTX *CTX, const char *message) {
  struct nt_node *node_header, *node_nt;
  struct nt_c c_nt, c_nt2;
  buffer *header, *body;
  char *decode, *x, *y;
  char heading[1024];
  int i = 0, spam_result = 0;

  CTX->_process_start = time(NULL);
                                                                                
  if (CTX->mode != DSM_PROCESS && (CTX->flags & DSF_SIGNATURE))
    return _ds_process_signature(CTX);

  header = buffer_create(NULL);
  body = buffer_create(NULL);

  if (header == NULL || body == NULL) { 
    LOG(LOG_CRIT, ERROR_MEM_ALLOC);
    buffer_destroy(header);
    buffer_destroy(body);
    return EUNKNOWN;
  }

  if (CTX->message == NULL && message != NULL) 
    CTX->message = _ds_actualize_message(message);  

  if (CTX->message == NULL) {
    LOG(LOG_WARNING, "_ds_actualize_message() failed");
    return EUNKNOWN;
  }

  /* Iterate through each component and create large header/body buffers */

  node_nt = c_nt_first(CTX->message->components, &c_nt);
  while(node_nt != NULL) {
    struct _ds_message_block *block = (struct _ds_message_block *) node_nt->ptr;
#ifdef VERBOSE
      LOGDEBUG("Processing component %d", i);
#endif
      if (block->headers == NULL || block->headers->items == 0) {
#ifdef VERBOSE
        LOGDEBUG("  : End of Message Identifier");
#endif

      /* Skip Attachments */
      } else if (block->media_type == MT_TEXT || block->media_type == MT_MESSAGE || i == 0) {

        /* Accumulate the headers */
        node_header = c_nt_first(block->headers, &c_nt2);
        while(node_header != NULL) {
          struct _ds_header_field *current_header = (struct _ds_header_field *) node_header->ptr;
          snprintf(heading, sizeof(heading), "%s: %s\n", current_header->heading, current_header->data);
          buffer_cat(header, heading);
          node_header = c_nt_next(block->headers, &c_nt2);
        }

        /* Accumulate the bodies */
        if (block->encoding == EN_BASE64 || block->encoding == EN_QUOTED_PRINTABLE) {
          LOGDEBUG("decoding message block from encoding type %d", block->encoding);
          decode = _ds_decode_block(block);
        } else
          decode = block->body->data;

        if (decode != NULL) {
          buffer_cat(body, decode);
          if (decode != block->body->data) {

            /* Switch the encoding */
            block->encoding = EN_7BIT;
            buffer_destroy(block->body);
            block->body = buffer_create(decode);
            free(decode);

            /* Rewrite the actual header */
            node_header = c_nt_first(block->headers, &c_nt2);
            while(node_header != NULL) {
              struct _ds_header_field *header = (struct _ds_header_field *) node_header->ptr;
              if (!strcasecmp(header->heading, "Content-Transfer-Encoding")) {
                free(header->data);
                header->data = strdup("7bit");
              }
              node_header = c_nt_next(block->headers, &c_nt2);
            }
          }
        }
      }
    node_nt = c_nt_next(CTX->message->components, &c_nt);
    i++;
  }

  if (CTX->flags & DSF_IGNOREHEADER)
    buffer_clear(header);

  if (header->data == NULL) 
    buffer_cat(header, " ");

  if (body->data == NULL) 
    buffer_cat(body, " ");

  /* -- PREFORMATTING BEGIN -- */

  /* Remove long HTML Comments (Including DSPAM statistics) */
  x = strstr(body->data, "<!--");
  while(x != NULL) {
    y = strstr(x, "-->");
    if (y != NULL) {
      memcpy(x, y+3, strlen(y+3)+1);
      x = strstr(x, "<!--");
    } else {
      x = strstr(x+4, "<!--");
    }
  }

  /* Remove short HTML Comments */
  x = strstr(body->data, "<!");
  while(x != NULL) {
    y = strstr(x, ">");
    if (y != NULL) {
      memcpy(x, y+1, strlen(y+1)+1);
      x = strstr(x, "<!");
    } else {
      x = strstr(x+2, "<!");
    }
  }

  /* Remove short html tags and useless tags */
  x = strstr(body->data, "<");
  while(x != NULL) {
    y = strstr(x, ">");
    if (y != NULL && (y-x<=15 || !strncmp(x+1, "td ", 3) || !strncmp(x+1, "table ", 6) || 
                                 !strncmp(x+1, "tr ", 3) || !strncmp(x+1, "div ", 4)   || 
                                 !strncmp(x+1, "p ", 2)  || !strncmp(x+1, "body ", 5)  ||
                                 !strchr(x, ' ') || strchr(x, ' ')>y)) {
      memcpy(x, y+1, strlen(y+1)+1);
      x = strstr(x, "<");
    } else {
      x = strstr(x+1, "<");
    }
  }

  /* Remove the multipart boundary */
/*
  if (boundaries->items>0) {
    node_nt = c_nt_first(boundaries, &c_nt);
    while(node_nt != NULL) {
      int len = strlen(node_nt->ptr);
      x = strstr(body->data, node_nt->ptr);
      while(x != NULL) {
        memset(x, 32, len);
        x = strstr(x+len, node_nt->ptr);
      }

      x = strstr(header->data, (char *)node_nt->ptr+2);
      while(x != NULL) {
        int len = strlen((char *)node_nt->ptr+2);
        memset(x, 32, len);
        x = strstr(x+len, (char *)node_nt->ptr+2);
      }
      node_nt = c_nt_next(boundaries, &c_nt);
    }
  }
*/

/*  nt_destroy(boundaries); */
  /* -- PREFORMATTING END -- */

  CTX->result = -1;
  spam_result = _ds_operate(CTX, header->data, body->data);

  if (spam_result == DSR_ISSPAM || spam_result == DSR_ISINNOCENT) {
    if (CTX->mode == DSM_FALSEPOSITIVE) 
      spam_result = DSR_ISINNOCENT;
    else if (CTX->mode == DSM_ADDSPAM) 
      spam_result = DSR_ISSPAM;
  }
 
  buffer_destroy(header);
  buffer_destroy(body);

  CTX->result = spam_result;

  if (CTX->result == DSR_ISSPAM || CTX->result == DSR_ISINNOCENT)
    return 0;
  else {
    LOGDEBUG("received invalid result (! DSR_ISSPAM || DSR_INNOCENT): %d\n", CTX->result)
    return EUNKNOWN;
  }
}

/* _ds_operate: operate on the message
                 - calculate the statistical probability the email is spam
                 - update tokens in dictionary according to result/mode
    parameters: DSPAM_CTX *CTX		pointer to context
                char *header		pointer to message header
                char *body		pointer to message body

       returns: DSR_ISSPAM		message is spam*
                DSR_ISINNOCENT		message is innocent*

       * returns result of actual calculation regardless of mode
*/

int _ds_operate(DSPAM_CTX *CTX, char *headers, char *body) {

  char *token;			/* current token */
  char joined_token[32];	/* used for de-obfuscating tokens */
  char *previous_token = NULL;  /* used for chained tokens */
	
  char *line = NULL;    /* header broken up into lines */
  char *url_body;       /* urls broken up */

  char heading[128];	/* current heading */
  int alloc_joined=0;	/* track joined token free()'s */
  int i;

  /* Bayesian Calculations */
  float bay_top = 0.0;	/* AB */
  float bay_bot = 0.0;	/* (1-A)(1-B) */

  /* Long Hashed Token Tree: Track tokens, frequencies, and stats */
  struct lht *freq = lht_create(1543);
  struct lht_node *node_lht;
  struct lht_c c_lht;
  struct _ds_spam_stat stat;

  struct tbt *index = tbt_create();	/* Binary tree index */
  struct tbt_node *node_tbt;

  struct nt *header;			/* header array */
  struct nt_node *node_nt;
  struct nt_c c_nt;

  joined_token[0] = 0;

  header = nt_create(NT_CHAR);

  if (freq == NULL || header == NULL || index == NULL) {
    tbt_destroy(index);
    nt_destroy(header);
    lht_destroy(freq);
    LOG(LOG_CRIT, "memory allocation failed");
    return EUNKNOWN;
  }

  CTX->result = (CTX->mode == DSM_ADDSPAM) ? 1 : 0;

  /* HEADER: Split up the text into tokens, include heading */
  line = strtok(headers, "\n");
  while(line != NULL) {
    nt_add(header, line);
    line = strtok(NULL, "\n");
  }

  node_nt = c_nt_first(header, &c_nt);
  heading[0] = 0;
  while(node_nt != NULL) {
    int is_received;
    joined_token[0] = 0;
    alloc_joined = 0;

    line = node_nt->ptr;
    token = strtok(line, ":");
    if (token != NULL && token[0] != 32 && token[0] != 9 && !strstr(token, " ")) {
      strlcpy(heading, token, 128);
      previous_token = NULL;
    }

#ifdef VERBOSE
    LOGDEBUG("Reading '%s' header", heading);
#endif

    is_received = (!strcmp(heading, "Received") ? 1 : 0);

    if (is_received)
      token = strtok(NULL, DELIMITERS_HEADING);
    else 
      token = strtok(NULL, DELIMITERS); 

    while(token != NULL) {
      int l;

      l  = strlen(token);
      if ((l>2 && l<25) || (l == 2 && !strchr(token, '$') && !strchr(token, '!'))) {

#ifdef VERBOSE
        LOGDEBUG("Processing '%s' token in '%s' header", token, heading);
#endif
 
        /* If we had to join a token together (e.g. S E X), process it */
        if (joined_token[0] != 0) {
          if (strlen(joined_token)<25 && joined_token[1]!=0) {
            if (! _ds_process_header_token(CTX, joined_token, previous_token, freq, heading) && (CTX->flags & DSF_CHAINED)) {
              alloc_joined = 1;
              previous_token = strdup(joined_token);
            }
          }
          joined_token[0] = 0;
        }

        if (! _ds_process_header_token(CTX, token, previous_token, freq, heading) && (CTX->flags & DSF_CHAINED)) {
          if (alloc_joined) {
            free(previous_token);
            alloc_joined = 0;
          }
          previous_token = token;
        }

      } else if (l==1 || (l==2 && (strchr(token, '$') || strchr(token, '!'))) ) {
        strlcat(joined_token, token, sizeof(joined_token));
      }

      if (is_received)
        token = strtok(NULL, DELIMITERS_HEADING);
      else
        token = strtok(NULL, DELIMITERS);
    }
    node_nt = c_nt_next(header, &c_nt);
 
    if (joined_token[0] != 0) {
      if (strlen(joined_token)<25 && joined_token[1]!=0) {
        _ds_process_header_token(CTX, joined_token, previous_token, freq, heading);
      }
    }

  }

  nt_destroy(header);

  if (alloc_joined) 
    free(previous_token);

  previous_token = NULL;

  /* BODY: Split up URLs into tokens, count frequency */
  url_body = strdup(body);
  if (url_body != NULL) {
    char combined_token[256];
    char *url_ptr = url_body;
    int url_length;
    unsigned long long crc;

    token = strstr(url_ptr, "http://");
    while(token != NULL) {
      url_ptr = token;

      token = strtok(token, " \n\">");
      if (token != NULL) {
        url_length = strlen(token);

        /* Individual tokens form the URL */
        token = strtok(token, DELIMITERS);
        while(token != NULL) {
          snprintf(combined_token, sizeof(combined_token), "Url*%s", token);
          crc = _ds_getcrc64(combined_token);
          lht_hit(freq, crc, combined_token);
          token = strtok(NULL, DELIMITERS);
        }

        memset(body+(url_ptr-url_body), 32, url_length);
        url_ptr += url_length + 1;
        token = strstr(url_ptr, "http://");
      } else 
        token = NULL;
    }
    free(url_body);
  }

  url_body = strdup(body);
  if (url_body != NULL) {
    char combined_token[256];
    char *url_ptr = url_body;
    int url_length;
    unsigned long long crc;

    url_ptr = url_body;
    token = strstr(url_ptr, "href=\"");
    while(token != NULL) {
      url_ptr = token+6;

      token = strtok(url_ptr, " \n\">");
      if (token != NULL) {
        url_length = strlen(token);

        /* Individual tokens form the URL */
        token = strtok(url_ptr, DELIMITERS);
        while(token != NULL) {
          snprintf(combined_token, sizeof(combined_token), "Url*%s", token);
          crc = _ds_getcrc64(combined_token);
          lht_hit(freq, crc, combined_token);
          token = strtok(NULL, DELIMITERS);
        }

        memset(body+(url_ptr-url_body), 32, url_length);

        url_ptr += url_length + 1;
        token = strstr(url_ptr, "href=\"");
      } else
        token = NULL;
    }

    free(url_body);
  }

  /* BODY: Split up the text into tokens, count frequency */
  joined_token[0] = 0;
  alloc_joined = 0;
  token = strtok(body, DELIMITERS);
  while(token != NULL) {
    int l = strlen(token);
    if ((l>2 && l<25) || (l == 2 && !strchr(token, '$') && !strchr(token, '!'))) {
      /* If we had to join a token together (e.g. S E X), process it */
      if (joined_token[0] != 0) {
        if (strlen(joined_token)<25 && joined_token[1]!=0) {
          if (! _ds_process_body_token(CTX, joined_token, previous_token, freq) && (CTX->flags & DSF_CHAINED)) {
            alloc_joined = 1;
            previous_token = strdup(joined_token);
          }
        }
        joined_token[0] = 0;
      }

      if (! _ds_process_body_token(CTX, token, previous_token, freq) && (CTX->flags & DSF_CHAINED)) {
        if (alloc_joined) {
          alloc_joined = 0;
          free(previous_token);
        }
        previous_token = token;
      }
    } else if (l==1 || (l==2 && (strchr(token, '$') || strchr(token, '!'))) ) {
      strlcat(joined_token, token, sizeof(joined_token));
    }
    token = strtok(NULL, DELIMITERS);
  }

  if (joined_token[0] != 0) {
    if (strlen(joined_token)<25 && joined_token[1]!=0) {
      _ds_process_body_token(CTX, joined_token, previous_token, freq);
    }
  }

  /* Load all token statistics */
  if (_ds_getall_spamrecords(CTX, freq)) {
    LOGDEBUG("_ds_getall_spamrecords() failed");
    return EUNKNOWN;
  }

  /* Create a binary tree index sorted by a token's delta from .5 */
  node_lht = c_lht_first(freq, &c_lht);
  while(node_lht != NULL) {
    _ds_calc_stat(CTX, node_lht->key, &node_lht->s);
    tbt_add(index, node_lht->s.probability, node_lht->key);
#ifdef VERBOSE
    LOGDEBUG("Token: %s [%f]", node_lht->token_name, node_lht->s.probability);
#endif

    node_lht = c_lht_next(freq, &c_lht);
  } 
 
  /* Take the 15 most interesting tokens and generate a score, update dictionary */

  if (index->items == 0) {
    LOGDEBUG("no tokens found in message");
    tbt_destroy(index);
    lht_destroy(freq);
    return EINVAL;
  }

  /* Initialize signature if requested */
  if (CTX->flags & DSF_SIGNATURE) {
    CTX->signature = calloc(1, sizeof(struct _ds_spam_signature));
    if (CTX->signature == NULL) {
      LOG(LOG_CRIT, "memory allocation error");
      tbt_destroy(index);
      lht_destroy(freq);
      return EUNKNOWN;
    }

    CTX->signature->length = sizeof(unsigned long long)*index->items;
    CTX->signature->data = malloc(CTX->signature->length);
    if (CTX->signature->data == NULL) {
      LOG(LOG_CRIT, "memory allocation error");
      free(CTX->signature);
      CTX->signature = NULL;
      tbt_destroy(index);
      lht_destroy(freq);
      return EUNKNOWN;
    }
  }

  node_tbt = tbt_first(index);

  CTX->probability = -1;

  for(i=0;i<index->items;i++) {
    unsigned long long crc;
    char *token_name;

    crc = node_tbt->token;
    token_name = lht_gettoken(freq, crc);

    if (CTX->flags & DSF_SIGNATURE) {
      memcpy((char *)CTX->signature->data + (i*sizeof(unsigned long long)), &crc, sizeof(unsigned long long));
    }

    if (lht_getspamstat(freq, crc, &stat) || token_name == NULL) {
      node_tbt = tbt_next(node_tbt);
      continue;
    }

    if (i<15) {
      LOGDEBUG("combining: [%2.6f] %s %ld %ld", stat.probability, token_name, stat.spam_hits, stat.innocent_hits);

      if (bay_top == 0.0) 
        bay_top = stat.probability;
      else
        bay_top *= stat.probability;

      if (bay_bot == 0.0)
        bay_bot = 1-stat.probability;
      else
        bay_bot *= (1-stat.probability);
    }

    if (i == 15 && CTX->probability < 0) 
      CTX->probability = (bay_top) / (bay_top + bay_bot);

    if (i>=15) {
      if (CTX->probability>=.9 && CTX->mode != DSM_FALSEPOSITIVE) 
        CTX->result = DSR_ISSPAM;

      if (CTX->result == DSR_ISSPAM) {
        stat.spam_hits++;
      } else if (CTX->mode != DSM_FALSEPOSITIVE) {
        stat.innocent_hits++;
      }

      if (CTX->mode == DSM_ADDSPAM && (! (CTX->flags & DSF_CORPUS))) 
        stat.innocent_hits -= (stat.innocent_hits>0) ? 1 : 0;

      if (CTX->mode == DSM_FALSEPOSITIVE) {
          stat.innocent_hits++;
          stat.spam_hits-= (stat.spam_hits>0) ? 1 : 0;
      }
    }

    lht_setspamstat(freq, crc, &stat);
    node_tbt = tbt_next(node_tbt);
  }

  CTX->probability = (bay_top) / (bay_top + bay_bot);

  if (CTX->probability>=.9 && CTX->mode != DSM_FALSEPOSITIVE) {
    CTX->result = DSR_ISSPAM;
  }

  if (CTX->result == DSR_ISSPAM) {
    CTX->totals.total_spam++;

    if (CTX->mode == DSM_ADDSPAM) {
      if (! (CTX->flags & DSF_CORPUS)) {
        CTX->totals.total_innocent--;
      }
      CTX->totals.spam_misses++;

      if (CTX->totals.total_innocent<0)
        CTX->totals.total_innocent = 0;
    }
  } else { 
    if (CTX->mode == DSM_FALSEPOSITIVE) { 
      CTX->totals.total_innocent++;
      CTX->totals.false_positives++;
      CTX->totals.total_spam--;
      if (CTX->totals.total_spam<0)
        CTX->totals.total_spam=0;
    } else {
      CTX->totals.total_innocent++;
    }
  }

  node_tbt = tbt_first(index);

  for(i=0;i<((index->items>15) ? 15 : index->items) && node_tbt != NULL;i++) {
    unsigned long long crc;
    char *token_name;

    crc = node_tbt->token;
    token_name = lht_gettoken(freq, crc);
    if (lht_getspamstat(freq, crc, &stat) || token_name == NULL) {
      node_tbt = tbt_next(node_tbt);
      continue;
    }

    if (CTX->result == DSR_ISSPAM) {
      stat.spam_hits++;
    } else if (CTX->mode != DSM_FALSEPOSITIVE) {
      stat.innocent_hits++;
    }

    if (CTX->mode == DSM_ADDSPAM && (! (CTX->flags & DSF_CORPUS)))
      stat.innocent_hits--;

    if (CTX->mode == DSM_FALSEPOSITIVE) {
      stat.innocent_hits++;

      stat.spam_hits--;
      if (stat.spam_hits<0) 
        stat.spam_hits = 0;
    }
    if (stat.innocent_hits<0)
      stat.innocent_hits = 0;
    
    lht_setspamstat(freq, crc, &stat);

    node_tbt = tbt_next(node_tbt);
  }

  if (_ds_setall_spamrecords(CTX, freq)) {
    LOGDEBUG("_ds_setall_spamrecords() failed");
    return EUNKNOWN;
  }
 
  tbt_destroy(index);
  lht_destroy(freq);

  if (CTX->mode == DSM_FALSEPOSITIVE) {
    CTX->probability = 0.0;
    CTX->result = DSR_ISINNOCENT;
  } else if (CTX->mode == DSM_ADDSPAM) {
    CTX->probability = 1.0;
    CTX->result = DSR_ISSPAM;
  }

  return (CTX->result == DSR_ISSPAM) ? DSR_ISSPAM : DSR_ISINNOCENT;
}

int _ds_process_signature(DSPAM_CTX *CTX) {
  unsigned long long token;
  int num_tokens, i;
  struct lht *freq = lht_create(1543);
  struct lht_node *node_lht;
  struct lht_c c_lht;

  if (CTX->signature == NULL) {
    LOGDEBUG("DSF_SIGNATURE specified, but no signature provided.");
    return EINVAL;
  }

  if (freq == NULL) {
    LOG(LOG_CRIT, ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }

  LOGDEBUG("processing signature.  length: %ld", CTX->signature->length);

  CTX->result = -1;

  if (CTX->mode == DSM_FALSEPOSITIVE) {
    CTX->totals.false_positives++;
    CTX->totals.total_spam -= (CTX->totals.total_spam>0) ? 1 : 0;
    CTX->totals.total_innocent++;
  } else {
    CTX->totals.spam_misses++;
    CTX->totals.total_spam++;
    CTX->totals.total_innocent -= (CTX->totals.total_innocent>0) ? 1 : 0;
  }

  num_tokens = CTX->signature->length / sizeof(unsigned long long);

  LOGDEBUG("reversing %d tokens", num_tokens);
  for(i=0;i<num_tokens;i++) {
    memcpy(&token, (char *)CTX->signature->data + (i*sizeof(unsigned long long)), sizeof(unsigned long long));
    lht_hit(freq, token, "-");
  }

  if (_ds_getall_spamrecords(CTX, freq)) 
    return EUNKNOWN;

  node_lht = c_lht_first(freq, &c_lht);
  while(node_lht != NULL) {
    if (CTX->mode == DSM_FALSEPOSITIVE) {
      node_lht->s.innocent_hits++;
      node_lht->s.spam_hits -= (node_lht->s.spam_hits>0) ? 1 : 0;
    } else {
      node_lht->s.innocent_hits -= (node_lht->s.innocent_hits>0) ? 1 : 0;
      node_lht->s.spam_hits++;
    }
    node_lht = c_lht_next(freq, &c_lht);
  }

  if (_ds_setall_spamrecords(CTX, freq))
    return EUNKNOWN;

  if (CTX->mode == DSM_FALSEPOSITIVE) {
    CTX->probability = 0.0;
    CTX->result = DSR_ISINNOCENT;
  } else {
    CTX->probability = 1.0;
    CTX->result = DSR_ISSPAM;
  }

  return 0;
}

/* load/save spam statistics */

int _ds_calc_stat(DSPAM_CTX *CTX, unsigned long long token, struct _ds_spam_stat *s) {
  int min_hits = 5;

  if (CTX->totals.total_innocent<500)
    min_hits = 20;

  if (CTX->totals.total_innocent<1000 && CTX->totals.total_innocent >=500) {
    float spams = (CTX->totals.total_spam*1.0/(CTX->totals.total_spam*1.0+CTX->totals.total_innocent*1.0))*100;
    if (spams>20)
      min_hits = 5+(spams/2);
  }

  if (CTX->mode == DSM_ADDSPAM)
    s->probability = .7;
  else
    s->probability = .4;

#ifdef BIAS
  s->probability = (s->spam_hits*1.0 / CTX->totals.total_spam*1.0) / ( (s->spam_hits*1.0 / CTX->totals.total_spam*1.0) + (s->innocent_hits*2.0 / CTX->totals.total_innocent*1.0) );
#else
  s->probability = (s->spam_hits*1.0 / CTX->totals.total_spam*1.0) / ( (s->spam_hits*1.0 / CTX->totals.total_spam*1.0) + (s->innocent_hits*1.0 / CTX->totals.total_innocent*1.0) );
#endif

  if (s->innocent_hits<0)
    s->innocent_hits = 0;

  if (s->spam_hits<0)
    s->spam_hits = 0;

  if (s->spam_hits == 0 && s->innocent_hits >0) {
    if (s->innocent_hits>50)
      s->probability = 0.0060;
    else if (s->innocent_hits>10) 
      s->probability = 0.0099;
    else 
      s->probability = 0.0100;
  } else if (s->spam_hits >0 && s->innocent_hits == 0) {
    if (s->spam_hits>=10)
      s->probability = 0.9901;
    else
      s->probability = 0.9900;
  }

#ifdef BIAS
  if (s->spam_hits + (2*s->innocent_hits)<min_hits || CTX->totals.total_innocent<min_hits)
#else
  if (s->spam_hits + s->innocent_hits<min_hits || CTX->totals.total_innocent<min_hits)
#endif
      s->probability = .4;

  if (s->probability < 0.0010)
    s->probability = 0.0010;

  if (s->probability > 0.9990) 
    s->probability = 0.9990;

  return 0;
}

int _ds_process_header_token (DSPAM_CTX *CTX, char *token, const char *previous_token, struct lht *freq, const char *heading) {
  int all_num = 1, i;
  int high_chars = 0;
  char combined_token[256];
  int len = 0;
  int is_received;
  unsigned long long crc;

  is_received = (!strcmp(heading, "Received") ? 1 : 0);

  if (is_received && strlen(token)<6) 
    return EINVAL;

  for(i=0;token[i]!=0;i++) {
    if (! isdigit((int) token[i]))
      all_num = 0;
    if (token[i]>=127 || iscntrl((int) token[i]))
      high_chars = 1;
  }

  len = i-1;

  if (isdigit((int) token[0])) {
    if ( token[len-1]!= '%')
      all_num = 1;
  }

  if (! isalnum((int) token[0]) && token[0] != '$' && token[0] != '#')
    all_num = 1;

  if (is_received)
    all_num = 0;

  /* Ignore tokens that are all numbers, or contain high ASCII characters */
  if (all_num || high_chars) 
     return EINVAL;

  /* Ignore 'Received', 'Date', and 'Message-Id' headers */

  if (strcasecmp(heading, "Date")) {
    if (heading[0] != 0)
      snprintf(combined_token, sizeof(combined_token), "%s*%s", heading, token);
    else
      strlcpy(combined_token, token, sizeof(combined_token));

    crc = _ds_getcrc64(combined_token);
#ifdef VERBOSE
    LOGDEBUG("Token Hit: '%s'", combined_token);
#endif
    lht_hit(freq, crc, combined_token);

    if ((CTX->flags & DSF_CHAINED) && previous_token != NULL && !is_received) {
     snprintf(combined_token, sizeof(combined_token), "%s*%s+%s", heading, previous_token, token);
      crc = _ds_getcrc64(combined_token);

      lht_hit(freq, crc, combined_token);
    }
  } else {
    return EINVAL;
  }

  return 0;
}

int _ds_process_body_token(DSPAM_CTX *CTX, char *token, const char *previous_token, struct lht *freq) {
  int all_num = 1, i;
  int high_chars = 0;
  char combined_token[256];
  int len;
  unsigned long long crc;

  for(i=0;token[i]!=0;i++) {
    if (! isdigit((int) token[i]))
      all_num = 0;
    if (token[i]>=127 || iscntrl((int) token[i]))
      high_chars = 1;
  }

  len = i-1;

  if (isdigit((int) token[0])) {
    int l = len - 1;
    if ( token[l]!= '%')
      all_num = 1;
  }

  if (! isalnum((int) token[0]) && token[0] != '$' && token[0] != '#')
    all_num = 1;

  /* Ignore tokens that are all numbers, or contain high ASCII characters */
  if (all_num || high_chars)
     return EINVAL;

  crc = _ds_getcrc64(token);

  lht_hit(freq, crc, token);

  if ((CTX->flags & DSF_CHAINED) && previous_token != NULL) {
   snprintf(combined_token, sizeof(combined_token), "%s+%s", previous_token, token);
     crc = _ds_getcrc64(combined_token);

     lht_hit(freq, crc,combined_token);
  }

  return 0;
}

int _ds_push_boundary(struct nt *stack, const char *boundary) {
  char *y = malloc(strlen(boundary)+3);
  if (y == NULL) 
    return EUNKNOWN;

  sprintf(y, "--%s", boundary);
  nt_add(stack, (char *) y);
  return 0;
}

char *_ds_pop_boundary(struct nt *stack) {
  struct nt_node *node, *last_node = NULL, *parent_node = NULL;
  struct nt_c c;
  char *boundary = NULL;

  node = c_nt_first(stack, &c);
  while(node != NULL) {
    parent_node = last_node;
    last_node = node;
    node = c_nt_next(stack, &c);
  }
  if (parent_node != NULL)
    parent_node->next = NULL;
  else
    stack->first = NULL;

  if (last_node == NULL) 
    return NULL;

  boundary = strdup(last_node->ptr);
  
  free(last_node->ptr);
  free(last_node);

  return boundary;
}

int _ds_match_boundary(struct nt *stack, const char *buff) {
  struct nt_node *node;
  struct nt_c c;

  node = c_nt_first(stack, &c);
  while(node != NULL) {
    if (!strncmp(buff, node->ptr, strlen(node->ptr))) { 
      return 1;
    }
    node = c_nt_next(stack, &c);
  }
  return 0;
}

unsigned long long _ds_getcrc64(const char *s) {
  static unsigned long long CRCTable[256];
  unsigned long long crc = 0;
  static int init = 0;
                                                                                                                                                                          
  if (!init) {
    int i;
    init = 1;
    for (i = 0; i <= 255; i++) {
      int j;
      unsigned long long part = i;
      for (j = 0; j < 8; j++) {
        if (part & 1)
          part = (part >> 1) ^ POLY64REV;
        else
          part >>= 1;
      }
      CRCTable[i] = part;
    }
  }
  for(;*s;*s++) {
    unsigned long long temp1 = crc >> 8;
    unsigned long long temp2 = CRCTable[(crc ^ (unsigned long long) *s) & 0xff];
    crc = temp1 ^ temp2;
  }
  return crc;
}

