/*
 DSPAM
 COPYRIGHT (C) 2003 NETWORK DWEEBS CORPORATION

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <error.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>

#include "language.h"
#include "config.h"
#include "util.h"

/* Compliments of Jay Freeman <saurik@saurik.com> */

#ifndef HAVE_STRSEP
char *strsep(char **stringp, const char *delim) {
    char *ret = *stringp;
    if (ret == NULL) return(NULL); 
    if ((*stringp = strpbrk(*stringp, delim)) != NULL) 
        *((*stringp)++) = '\0';
    return(ret);
}
#endif

/*
    Subroutine: chomp
   Description: removes a trailing newline character
    Parameters: char *string    string to chomp
*/

void chomp(char *string) {
  int len;
  if (string == NULL) return;
  len  = strlen(string);
  if (len == 0) return;
  if (string[len-1] == 10) {
    string[len-1] = 0;
    len--;
  }
  if (string[len-1] == 13)
    string[len-1] = 0;
  return;
}

#ifndef HAVE_STRLCPY
/*
 * Appends src to string dst of size siz (unlike strncat, siz is the
 * full size of dst, not space left).  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz <= strlen(dst)).
 * Returns strlen(src) + MIN(siz, strlen(initial dst)).
 * If retval >= siz, truncation occurred.
 */
size_t
strlcat(dst, src, siz)
        char *dst;
        const char *src;
        size_t siz;
{
        register char *d = dst;
        register const char *s = src;
        register size_t n = siz;
        size_t dlen;

        /* Find the end of dst and adjust bytes left but don't go past end */
        while (n-- != 0 && *d != '\0')
                d++;
        dlen = d - dst;
        n = siz - dlen;

        if (n == 0)
                return(dlen + strlen(s));
        while (*s != '\0') {
                if (n != 1) {
                        *d++ = *s;
                        n--;
                }
                s++;
        }
        *d = '\0';

        return(dlen + (s - src));       /* count does not include NUL */
}

/*
 * Copy src to string dst of size siz.  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
size_t
strlcpy(dst, src, siz)
        char *dst;
        const char *src;
        size_t siz;
{
        register char *d = dst;
        register const char *s = src;
        register size_t n = siz;

        /* Copy as many bytes as will fit */
        if (n != 0 && --n != 0) {
                do {
                        if ((*d++ = *s++) == 0)
                                break;
                } while (--n != 0);
        }

        /* Not enough room in dst, add NUL and traverse rest of src */
        if (n == 0) {
                if (siz != 0)
                        *d = '\0';              /* NUL-terminate dst */
                while (*s++)
                        ;
        }

        return(s - src - 1);    /* count does not include NUL */
}
#endif

const char *_ds_userdir_path(const char *filename, const char *extension) {
  static char path[MAX_FILENAME_LENGTH];
                                                                                                                                                                          
  if (filename==NULL || extension==NULL || filename[0] == 0) {
    path[0] = 0;
    return path;
   }

#ifdef LARGE_SCALE
   /* Locks use USERDIR */
   if (!strcmp(extension, "lock")) {
     snprintf(path, sizeof(path), "%s/%s.%s", USERDIR, filename, extension);
     return path;
   }
#endif

#ifdef HOMEDIR
  if (!strcmp(extension, "nodspam") || !strcmp(extension, "dspam")) {
    struct passwd *p;
    p = getpwnam(filename);
    if (p != NULL) {
      snprintf(path, sizeof(path), "%s/.%s", p->pw_dir, extension);
#ifdef DEBUG
      LOGDEBUG("using %s as path for dotfile", path);
#endif
      return path;
    }
  }
#endif /* HOMEDIR */
                                                                                                                                                                          
#ifdef LARGESCALE
  if (filename[1]!=0) {
    if (extension[0] == 0)
      snprintf(path, MAX_FILENAME_LENGTH, "%s/%c/%c/%s", USERDIR, filename[0], filename[1], filename);
    else
      snprintf(path, MAX_FILENAME_LENGTH, "%s/%c/%c/%s.%s", USERDIR, filename[0], filename[1], filename, extension);
  } else {
    if (extension[0] == 0)
      snprintf(path, MAX_FILENAME_LENGTH, "%s/%c/%s", USERDIR, filename[0], filename);
    else
      snprintf(path, MAX_FILENAME_LENGTH, "%s/%c/%s.%s", USERDIR, filename[0], filename, extension);
  }
#else
  snprintf(path, MAX_FILENAME_LENGTH, "%s/%s.%s", USERDIR, filename, extension);
#endif
                                                                                                                                                                          
  return path;
}

